import abc
import warnings
from functools import wraps
from types import ModuleType
from typing import Literal

import numpy as np

from yt.config import ytcfg
from yt.data_objects.image_array import ImageArray
from yt.funcs import ensure_numpy_array, is_sequence, mylog
from yt.geometry.grid_geometry_handler import GridIndex
from yt.geometry.oct_geometry_handler import OctreeIndex
from yt.utilities.amr_kdtree.api import AMRKDTree
from yt.utilities.configure import YTConfig, configuration_callbacks
from yt.utilities.lib.bounding_volume_hierarchy import BVH
from yt.utilities.lib.misc_utilities import zlines, zpoints
from yt.utilities.lib.octree_raytracing import OctreeRayTracing
from yt.utilities.lib.partitioned_grid import PartitionedGrid
from yt.utilities.on_demand_imports import NotAModule
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    ParallelAnalysisInterface,
)
from yt.visualization.image_writer import apply_colormap

from .transfer_function_helper import TransferFunctionHelper
from .transfer_functions import (
    ColorTransferFunction,
    ProjectionTransferFunction,
    TransferFunction,
)
from .utils import (
    data_source_or_all,
    get_corners,
    new_interpolated_projection_sampler,
    new_mesh_sampler,
    new_projection_sampler,
    new_volume_render_sampler,
)
from .zbuffer_array import ZBuffer

OptionalModule = ModuleType | NotAModule
mesh_traversal: OptionalModule = NotAModule("pyembree")
mesh_construction: OptionalModule = NotAModule("pyembree")


def set_raytracing_engine(
    engine: Literal["yt", "embree"],
) -> None:
    """
    Safely switch raytracing engines at runtime.

    Parameters
    ----------

    engine: 'yt' or 'embree'
      - 'yt' selects the default engine.
      - 'embree' requires extra installation steps, see
        https://yt-project.org/doc/visualizing/unstructured_mesh_rendering.html?highlight=pyembree#optional-embree-installation

    Raises
    ------

    UserWarning
      Raised if the required engine is not available.
      In this case, the default engine is restored.

    """
    from yt.config import ytcfg

    global mesh_traversal, mesh_construction

    if engine == "embree":
        try:
            from yt.utilities.lib.embree_mesh import (  # type: ignore
                mesh_construction,
                mesh_traversal,
            )
        except (ImportError, ValueError) as exc:
            # Catch ValueError in case size of objects in Cython change
            warnings.warn(
                "Failed to switch to embree raytracing engine. "
                f"The following error was raised:\n{exc}",
                stacklevel=2,
            )
            mesh_traversal = NotAModule("pyembree")
            mesh_construction = NotAModule("pyembree")
            ytcfg["yt", "ray_tracing_engine"] = "yt"
        else:
            ytcfg["yt", "ray_tracing_engine"] = "embree"
    else:
        mesh_traversal = NotAModule("pyembree")
        mesh_construction = NotAModule("pyembree")
        ytcfg["yt", "ray_tracing_engine"] = "yt"


def _init_raytracing_engine(ytcfg: YTConfig) -> None:
    # validate option from configuration file or fall back to default engine
    set_raytracing_engine(engine=ytcfg["yt", "ray_tracing_engine"])


configuration_callbacks.append(_init_raytracing_engine)


def invalidate_volume(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        ret = f(*args, **kwargs)
        obj = args[0]
        if isinstance(obj._transfer_function, ProjectionTransferFunction):
            obj.sampler_type = "projection"
            obj._log_field = False
            obj._use_ghost_zones = False
        del obj.volume
        obj._volume_valid = False
        return ret

    return wrapper


def validate_volume(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        obj = args[0]
        fields = [obj.field]
        log_fields = [obj.log_field]
        if obj.weight_field is not None:
            fields.append(obj.weight_field)
            log_fields.append(obj.log_field)
        if not obj._volume_valid:
            obj.volume.set_fields(
                fields, log_fields, no_ghost=(not obj.use_ghost_zones)
            )
        obj._volume_valid = True
        return f(*args, **kwargs)

    return wrapper


class RenderSource(ParallelAnalysisInterface, abc.ABC):
    """Base Class for Render Sources.

    Will be inherited for volumes, streamlines, etc.

    """

    volume_method: str | None = None

    def __init__(self):
        super().__init__()
        self.opaque = False
        self.zbuffer = None

    @abc.abstractmethod
    def render(self, camera, zbuffer=None):
        pass

    @abc.abstractmethod
    def _validate(self):
        pass


class OpaqueSource(RenderSource):
    """A base class for opaque render sources.

    Will be inherited from for LineSources, BoxSources, etc.

    """

    def __init__(self):
        super().__init__()
        self.opaque = True

    def set_zbuffer(self, zbuffer):
        self.zbuffer = zbuffer


def create_volume_source(data_source, field):
    data_source = data_source_or_all(data_source)
    ds = data_source.ds
    index_class = ds.index.__class__
    if issubclass(index_class, GridIndex):
        return KDTreeVolumeSource(data_source, field)
    elif issubclass(index_class, OctreeIndex):
        return OctreeVolumeSource(data_source, field)
    else:
        raise NotImplementedError


class VolumeSource(RenderSource, abc.ABC):
    """A class for rendering data from a volumetric data source

    Examples of such sources include a sphere, cylinder, or the
    entire computational domain.

    A :class:`VolumeSource` provides the framework to decompose an arbitrary
    yt data source into bricks that can be traversed and volume rendered.

    Parameters
    ----------
    data_source: :class:`AMR3DData` or :class:`Dataset`, optional
        This is the source to be rendered, which can be any arbitrary yt
        data object or dataset.
    field : string
        The name of the field to be rendered.

    Examples
    --------

    The easiest way to make a VolumeSource is to use the volume_render
    function, so that the VolumeSource gets created automatically. This
    example shows how to do this and then access the resulting source:

    >>> import yt
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> im, sc = yt.volume_render(ds)
    >>> volume_source = sc.get_source(0)

    You can also create VolumeSource instances by hand and add them to Scenes.
    This example manually creates a VolumeSource, adds it to a scene, sets the
    camera, and renders an image.

    >>> import yt
    >>> from yt.visualization.volume_rendering.api import (
    ...     Camera, Scene, create_volume_source)
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> sc = Scene()
    >>> source = create_volume_source(ds.all_data(), "density")
    >>> sc.add_source(source)
    >>> sc.add_camera()
    >>> im = sc.render()

    """

    _image = None
    data_source = None

    def __init__(self, data_source, field):
        r"""Initialize a new volumetric source for rendering."""
        super().__init__()
        self.data_source = data_source_or_all(data_source)
        field = self.data_source._determine_fields(field)[0]
        self.current_image = None
        self.check_nans = False
        self.num_threads = 0
        self.num_samples = 10
        self.sampler_type = "volume-render"

        self._volume_valid = False

        # these are caches for properties, defined below
        self._volume = None
        self._transfer_function = None
        self._field = field
        self._log_field = self.data_source.ds.field_info[field].take_log
        self._use_ghost_zones = False
        self._weight_field = None

        self.tfh = TransferFunctionHelper(self.data_source.pf)
        self.tfh.set_field(self.field)

    @property
    def transfer_function(self):
        """The transfer function associated with this VolumeSource"""
        if self._transfer_function is not None:
            return self._transfer_function

        if self.tfh.tf is not None:
            self._transfer_function = self.tfh.tf
            return self._transfer_function

        mylog.info("Creating transfer function")
        self.tfh.set_field(self.field)
        self.tfh.set_log(self.log_field)
        self.tfh.build_transfer_function()
        self.tfh.setup_default()
        self._transfer_function = self.tfh.tf

        return self._transfer_function

    @transfer_function.setter
    def transfer_function(self, value):
        self.tfh.tf = None
        valid_types = (
            TransferFunction,
            ColorTransferFunction,
            ProjectionTransferFunction,
            type(None),
        )
        if not isinstance(value, valid_types):
            raise RuntimeError(
                "transfer_function not a valid type, "
                f"received object of type {type(value)}"
            )
        if isinstance(value, ProjectionTransferFunction):
            self.sampler_type = "projection"
            if self._volume is not None:
                fields = [self.field]
                if self.weight_field is not None:
                    fields.append(self.weight_field)
                self._volume_valid = False
        self._transfer_function = value

    @property
    def volume(self):
        """The abstract volume associated with this VolumeSource

        This object does the heavy lifting to access data in an efficient manner
        using a KDTree
        """
        return self._get_volume()

    @volume.setter
    def volume(self, value):
        assert isinstance(value, AMRKDTree)
        del self._volume
        self._field = value.fields
        self._log_field = value.log_fields
        self._volume = value
        assert self._volume_valid

    @volume.deleter
    def volume(self):
        del self._volume
        self._volume = None

    @property
    def field(self):
        """The field to be rendered"""
        return self._field

    @field.setter
    @invalidate_volume
    def field(self, value):
        field = self.data_source._determine_fields(value)
        if len(field) > 1:
            raise RuntimeError(
                "VolumeSource.field can only be a single field but received "
                f"multiple fields: {field}"
            )
        field = field[0]
        if self._field != field:
            log_field = self.data_source.ds.field_info[field].take_log
            self.tfh.bounds = None
        else:
            log_field = self._log_field
        self._log_field = log_field
        self._field = value
        self.transfer_function = None
        self.tfh.set_field(value)
        self.tfh.set_log(log_field)

    @property
    def log_field(self):
        """Whether or not the field rendering is computed in log space"""
        return self._log_field

    @log_field.setter
    @invalidate_volume
    def log_field(self, value):
        self.transfer_function = None
        self.tfh.set_log(value)
        self._log_field = value

    @property
    def use_ghost_zones(self):
        """Whether or not ghost zones are used to estimate vertex-centered data
        values at grid boundaries"""
        return self._use_ghost_zones

    @use_ghost_zones.setter
    @invalidate_volume
    def use_ghost_zones(self, value):
        self._use_ghost_zones = value

    @property
    def weight_field(self):
        """The weight field for the rendering

        Currently this is only used for off-axis projections.
        """
        return self._weight_field

    @weight_field.setter
    @invalidate_volume
    def weight_field(self, value):
        self._weight_field = value

    def set_transfer_function(self, transfer_function):
        """Set transfer function for this source"""
        self.transfer_function = transfer_function
        return self

    def _validate(self):
        """Make sure that all dependencies have been met"""
        if self.data_source is None:
            raise RuntimeError("Data source not initialized")

    def set_volume(self, volume):
        """Associates an AMRKDTree with the VolumeSource"""
        self.volume = volume
        return self

    def set_field(self, field):
        """Set the source's field to render

        Parameters
        ----------

        field: field name
            The field to render
        """
        self.field = field
        return self

    def set_log(self, log_field):
        """Set whether the rendering of the source's field is done in log space

        Generally volume renderings of data whose values span a large dynamic
        range should be done on log space and volume renderings of data with
        small dynamic range should be done in linear space.

        Parameters
        ----------

        log_field: boolean
            If True, the volume rendering will be done in log space, and if False
            will be done in linear space.
        """
        self.log_field = log_field
        return self

    def set_weight_field(self, weight_field):
        """Set the source's weight field

        .. note::

          This is currently only used for renderings using the
          ProjectionTransferFunction

        Parameters
        ----------

        weight_field: field name
            The weight field to use in the rendering
        """
        self.weight_field = weight_field
        return self

    def set_use_ghost_zones(self, use_ghost_zones):
        """Set whether or not interpolation at grid edges uses ghost zones

        Parameters
        ----------

        use_ghost_zones: boolean
            If True, the AMRKDTree estimates vertex centered data using ghost
            zones, which can eliminate seams in the resulting volume rendering.
            Defaults to False for performance reasons.

        """
        self.use_ghost_zones = use_ghost_zones
        return self

    def set_sampler(self, camera, interpolated=True):
        """Sets a volume render sampler

        The type of sampler is determined based on the ``sampler_type`` attribute
        of the VolumeSource. Currently the ``volume_render`` and ``projection``
        sampler types are supported.

        The 'interpolated' argument is only meaningful for projections. If True,
        the data is first interpolated to the cell vertices, and then
        tri-linearly interpolated to the ray sampling positions. If False, then
        the cell-centered data is simply accumulated along the
        ray. Interpolation is always performed for volume renderings.

        """
        if self.sampler_type == "volume-render":
            sampler = new_volume_render_sampler(camera, self)
        elif self.sampler_type == "projection" and interpolated:
            sampler = new_interpolated_projection_sampler(camera, self)
        elif self.sampler_type == "projection":
            sampler = new_projection_sampler(camera, self)
        else:
            NotImplementedError(f"{self.sampler_type} not implemented yet")
        self.sampler = sampler
        assert self.sampler is not None

    @abc.abstractmethod
    def _get_volume(self):
        """The abstract volume associated with this VolumeSource

        This object does the heavy lifting to access data in an efficient manner
        using a KDTree
        """
        pass

    @abc.abstractmethod
    @validate_volume
    def render(self, camera, zbuffer=None):
        """Renders an image using the provided camera

        Parameters
        ----------
        camera: :class:`yt.visualization.volume_rendering.camera.Camera` instance
            A volume rendering camera. Can be any type of camera.
        zbuffer: :class:`yt.visualization.volume_rendering.zbuffer_array.Zbuffer` instance  # noqa: E501
            A zbuffer array. This is used for opaque sources to determine the
            z position of the source relative to other sources. Only useful if
            you are manually calling render on multiple sources. Scene.render
            uses this internally.

        Returns
        -------
        A :class:`yt.data_objects.image_array.ImageArray` instance containing
        the rendered image.

        """
        pass

    def finalize_image(self, camera, image):
        """Parallel reduce the image.

        Parameters
        ----------
        camera: :class:`yt.visualization.volume_rendering.camera.Camera` instance
            The camera used to produce the volume rendering image.
        image: :class:`yt.data_objects.image_array.ImageArray` instance
            A reference to an image to fill
        """
        image.shape = camera.resolution[0], camera.resolution[1], 4
        # If the call is from VR, the image is rotated by 180 to get correct
        # up direction
        if not self.transfer_function.grey_opacity:
            image[:, :, 3] = 1
        return image

    def __repr__(self):
        disp = f"<Volume Source>:{str(self.data_source)} "
        disp += f"transfer_function:{str(self._transfer_function)}"
        return disp


class KDTreeVolumeSource(VolumeSource):
    volume_method = "KDTree"

    def _get_volume(self):
        """The abstract volume associated with this VolumeSource

        This object does the heavy lifting to access data in an efficient manner
        using a KDTree
        """

        if self._volume is None:
            mylog.info("Creating volume")
            volume = AMRKDTree(self.data_source.ds, data_source=self.data_source)
            self._volume = volume

        return self._volume

    @validate_volume
    def render(self, camera, zbuffer=None):
        """Renders an image using the provided camera

        Parameters
        ----------
        camera: :class:`yt.visualization.volume_rendering.camera.Camera`
            A volume rendering camera. Can be any type of camera.
        zbuffer: :class:`yt.visualization.volume_rendering.zbuffer_array.Zbuffer`
            A zbuffer array. This is used for opaque sources to determine the
            z position of the source relative to other sources. Only useful if
            you are manually calling render on multiple sources. Scene.render
            uses this internally.

        Returns
        -------
        A :class:`yt.data_objects.image_array.ImageArray` containing
        the rendered image.

        """
        self.zbuffer = zbuffer
        self.set_sampler(camera)
        assert self.sampler is not None

        mylog.debug("Casting rays")
        total_cells = 0
        if self.check_nans:
            for brick in self.volume.bricks:
                for data in brick.my_data:
                    if np.any(np.isnan(data)):
                        raise RuntimeError

        for brick in self.volume.traverse(camera.lens.viewpoint):
            mylog.debug("Using sampler %s", self.sampler)
            self.sampler(brick, num_threads=self.num_threads)
            total_cells += np.prod(brick.my_data[0].shape)
        mylog.debug("Done casting rays")
        self.current_image = self.finalize_image(camera, self.sampler.aimage)

        if zbuffer is None:
            self.zbuffer = ZBuffer(
                self.current_image, np.full(self.current_image.shape[:2], np.inf)
            )

        return self.current_image

    def finalize_image(self, camera, image):
        if self._volume is not None:
            image = self.volume.reduce_tree_images(
                image,
                camera.lens.viewpoint,
                use_opacity=self.transfer_function.grey_opacity,
            )

        return super().finalize_image(camera, image)


class OctreeVolumeSource(VolumeSource):
    volume_method = "Octree"

    def __init__(self, *args, **kwa):
        super().__init__(*args, **kwa)
        self.set_use_ghost_zones(True)

    def _get_volume(self):
        """The abstract volume associated with this VolumeSource

        This object does the heavy lifting to access data in an efficient manner
        using an octree.
        """

        if self._volume is None:
            mylog.info("Creating volume")
            volume = OctreeRayTracing(self.data_source)
            self._volume = volume

        return self._volume

    @validate_volume
    def render(self, camera, zbuffer=None):
        """Renders an image using the provided camera

        Parameters
        ----------
        camera: :class:`yt.visualization.volume_rendering.camera.Camera` instance
            A volume rendering camera. Can be any type of camera.
        zbuffer: :class:`yt.visualization.volume_rendering.zbuffer_array.Zbuffer` instance  # noqa: E501
            A zbuffer array. This is used for opaque sources to determine the
            z position of the source relative to other sources. Only useful if
            you are manually calling render on multiple sources. Scene.render
            uses this internally.

        Returns
        -------
        A :class:`yt.data_objects.image_array.ImageArray` instance containing
        the rendered image.

        """
        self.zbuffer = zbuffer
        self.set_sampler(camera)
        if self.sampler is None:
            raise RuntimeError(
                "No sampler set. This is likely a bug as it should never happen."
            )

        data = self.data_source

        dx = data["dx"].to_value("unitary")[:, None]
        xyz = np.stack([data[_].to_value("unitary") for _ in "xyz"], axis=-1)
        LE = xyz - dx / 2
        RE = xyz + dx / 2

        mylog.debug("Gathering data")
        dt = np.stack(list(self.volume.data) + [*LE.T, *RE.T], axis=-1).reshape(
            1, len(dx), 14, 1
        )
        mask = np.full(dt.shape[1:], 1, dtype=np.uint8)
        dims = np.array([1, 1, 1], dtype="int64")
        pg = PartitionedGrid(0, dt, mask, LE.flatten(), RE.flatten(), dims, n_fields=1)

        mylog.debug("Casting rays")
        self.sampler(pg, oct=self.volume.octree)
        mylog.debug("Done casting rays")

        self.current_image = self.finalize_image(camera, self.sampler.aimage)

        if zbuffer is None:
            self.zbuffer = ZBuffer(
                self.current_image, np.full(self.current_image.shape[:2], np.inf)
            )

        return self.current_image


class MeshSource(OpaqueSource):
    """A source for unstructured mesh data.

    This functionality requires the embree ray-tracing engine and the
    associated pyembree python bindings to be installed in order to
    function.

    A :class:`MeshSource` provides the framework to volume render
    unstructured mesh data.

    Parameters
    ----------
    data_source: :class:`AMR3DData` or :class:`Dataset`, optional
        This is the source to be rendered, which can be any arbitrary yt
        data object or dataset.
    field : string
        The name of the field to be rendered.

    Examples
    --------
    >>> source = MeshSource(ds, ("connect1", "convected"))
    """

    _image = None
    data_source = None

    def __init__(self, data_source, field):
        r"""Initialize a new unstructured mesh source for rendering."""
        super().__init__()
        self.data_source = data_source_or_all(data_source)
        field = self.data_source._determine_fields(field)[0]
        self.field = field
        self.volume = None
        self.current_image = None
        self.engine = ytcfg.get("yt", "ray_tracing_engine")

        # default color map
        self._cmap = ytcfg.get("yt", "default_colormap")
        self._color_bounds = None

        # default mesh annotation options
        self._annotate_mesh = False
        self._mesh_line_color = None
        self._mesh_line_alpha = 1.0

        # Error checking
        assert self.field is not None
        assert self.data_source is not None
        if self.field[0] == "all":
            raise NotImplementedError(
                "Mesh unions are not implemented for 3D rendering"
            )

        if self.engine == "embree":
            self.volume = mesh_traversal.YTEmbreeScene()
            self.build_volume_embree()
        elif self.engine == "yt":
            self.build_volume_bvh()
        else:
            raise NotImplementedError(
                "Invalid ray-tracing engine selected. Choices are 'embree' and 'yt'."
            )

    @property
    def cmap(self):
        """
        This is the name of the colormap that will be used when rendering
        this MeshSource object. Should be a string, like 'cmyt.arbre', or 'cmyt.dusk'.

        """
        return self._cmap

    @cmap.setter
    def cmap(self, cmap_name):
        self._cmap = cmap_name
        if hasattr(self, "data"):
            self.current_image = self.apply_colormap()

    @property
    def color_bounds(self):
        """
        These are the bounds that will be used with the colormap to the display
        the rendered image. Should be a (vmin, vmax) tuple, like (0.0, 2.0). If
        None, the bounds will be automatically inferred from the max and min of
        the rendered data.

        """
        return self._color_bounds

    @color_bounds.setter
    def color_bounds(self, bounds):
        self._color_bounds = bounds
        if hasattr(self, "data"):
            self.current_image = self.apply_colormap()

    def _validate(self):
        """Make sure that all dependencies have been met"""
        if self.data_source is None:
            raise RuntimeError("Data source not initialized.")

        if self.volume is None:
            raise RuntimeError("Volume not initialized.")

    def build_volume_embree(self):
        """

        This constructs the mesh that will be ray-traced by pyembree.

        """
        ftype, fname = self.field
        mesh_id = int(ftype[-1]) - 1
        index = self.data_source.ds.index
        offset = index.meshes[mesh_id]._index_offset
        field_data = self.data_source[self.field].d  # strip units

        vertices = index.meshes[mesh_id].connectivity_coords
        indices = index.meshes[mesh_id].connectivity_indices - offset

        # if this is an element field, promote to 2D here
        if len(field_data.shape) == 1:
            field_data = np.expand_dims(field_data, 1)

        # Here, we decide whether to render based on high-order or
        # low-order geometry. Right now, high-order geometry is only
        # implemented for 20-point hexes.
        if indices.shape[1] == 20 or indices.shape[1] == 10:
            self.mesh = mesh_construction.QuadraticElementMesh(
                self.volume, vertices, indices, field_data
            )
        else:
            # if this is another type of higher-order element, we demote
            # to 1st order here, for now.
            if indices.shape[1] == 27:
                # hexahedral
                mylog.warning("27-node hexes not yet supported, dropping to 1st order.")
                field_data = field_data[:, 0:8]
                indices = indices[:, 0:8]

            self.mesh = mesh_construction.LinearElementMesh(
                self.volume, vertices, indices, field_data
            )

    def build_volume_bvh(self):
        """

        This constructs the mesh that will be ray-traced.

        """
        ftype, fname = self.field
        mesh_id = int(ftype[-1]) - 1
        index = self.data_source.ds.index
        offset = index.meshes[mesh_id]._index_offset
        field_data = self.data_source[self.field].d  # strip units

        vertices = index.meshes[mesh_id].connectivity_coords
        indices = index.meshes[mesh_id].connectivity_indices - offset

        # if this is an element field, promote to 2D here
        if len(field_data.shape) == 1:
            field_data = np.expand_dims(field_data, 1)

        # Here, we decide whether to render based on high-order or
        # low-order geometry.
        if indices.shape[1] == 27:
            # hexahedral
            mylog.warning("27-node hexes not yet supported, dropping to 1st order.")
            field_data = field_data[:, 0:8]
            indices = indices[:, 0:8]

        self.volume = BVH(vertices, indices, field_data)

    def render(self, camera, zbuffer=None):
        """Renders an image using the provided camera

        Parameters
        ----------
        camera: :class:`yt.visualization.volume_rendering.camera.Camera`
            A volume rendering camera. Can be any type of camera.
        zbuffer: :class:`yt.visualization.volume_rendering.zbuffer_array.Zbuffer`
            A zbuffer array. This is used for opaque sources to determine the
            z position of the source relative to other sources. Only useful if
            you are manually calling render on multiple sources. Scene.render
            uses this internally.

        Returns
        -------
        A :class:`yt.data_objects.image_array.ImageArray` containing
        the rendered image.

        """

        shape = (camera.resolution[0], camera.resolution[1], 4)
        if zbuffer is None:
            empty = np.empty(shape, dtype="float64")
            z = np.empty(empty.shape[:2], dtype="float64")
            empty[:] = 0.0
            z[:] = np.inf
            zbuffer = ZBuffer(empty, z)
        elif zbuffer.rgba.shape != shape:
            zbuffer = ZBuffer(zbuffer.rgba.reshape(shape), zbuffer.z.reshape(shape[:2]))
        self.zbuffer = zbuffer

        self.sampler = new_mesh_sampler(camera, self, engine=self.engine)

        mylog.debug("Casting rays")
        self.sampler(self.volume)
        mylog.debug("Done casting rays")

        self.finalize_image(camera)
        self.current_image = self.apply_colormap()

        zbuffer += ZBuffer(self.current_image.astype("float64"), self.sampler.azbuffer)
        zbuffer.rgba = ImageArray(zbuffer.rgba)
        self.zbuffer = zbuffer
        self.current_image = self.zbuffer.rgba

        if self._annotate_mesh:
            self.current_image = self.annotate_mesh_lines(
                self._mesh_line_color, self._mesh_line_alpha
            )

        return self.current_image

    def finalize_image(self, camera):
        sam = self.sampler

        # reshape data
        Nx = camera.resolution[0]
        Ny = camera.resolution[1]
        self.data = sam.aimage[:, :, 0].reshape(Nx, Ny)

    def annotate_mesh_lines(self, color=None, alpha=1.0):
        r"""

        Modifies this MeshSource by drawing the mesh lines.
        This modifies the current image by drawing the element
        boundaries and returns the modified image.

        Parameters
        ----------
        color: array_like of shape (4,), optional
            The RGBA value to use to draw the mesh lines.
            Default is black.
        alpha : float, optional
            The opacity of the mesh lines. Default is 255 (solid).

        """

        self.annotate_mesh = True
        self._mesh_line_color = color
        self._mesh_line_alpha = alpha

        if color is None:
            color = np.array([0, 0, 0, alpha])

        locs = (self.sampler.amesh_lines == 1,)

        self.current_image[:, :, 0][locs] = color[0]
        self.current_image[:, :, 1][locs] = color[1]
        self.current_image[:, :, 2][locs] = color[2]
        self.current_image[:, :, 3][locs] = color[3]

        return self.current_image

    def apply_colormap(self):
        """

        Applies a colormap to the current image without re-rendering.

        Returns
        -------
        current_image : A new image with the specified color scale applied to
            the underlying data.


        """

        image = (
            apply_colormap(
                self.data, color_bounds=self._color_bounds, cmap_name=self._cmap
            )
            / 255.0
        )
        alpha = image[:, :, 3]
        alpha[self.sampler.aimage_used == -1] = 0.0
        image[:, :, 3] = alpha
        return image

    def __repr__(self):
        disp = f"<Mesh Source>:{str(self.data_source)} "
        return disp


class PointSource(OpaqueSource):
    r"""A rendering source of opaque points in the scene.

    This class provides a mechanism for adding points to a scene; these
    points will be opaque, and can also be colored.

    Parameters
    ----------
    positions: array_like of shape (N, 3)
        The positions of points to be added to the scene. If specified with no
        units, the positions will be assumed to be in code units.
    colors : array_like of shape (N, 4), optional
        The colors of the points, including an alpha channel, in floating
        point running from 0..1.
    color_stride : int, optional
        The stride with which to access the colors when putting them on the
        scene.
    radii : array_like of shape (N), optional
        The radii of the points in the final image, in pixels (int)

    Examples
    --------

    This example creates a volume rendering and adds 1000 random points to
    the image:

    >>> import yt
    >>> import numpy as np
    >>> from yt.visualization.volume_rendering.api import PointSource
    >>> from yt.units import kpc
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

    >>> im, sc = yt.volume_render(ds)

    >>> npoints = 1000
    >>> vertices = np.random.random([npoints, 3]) * 1000 * kpc
    >>> colors = np.random.random([npoints, 4])
    >>> colors[:, 3] = 1.0

    >>> points = PointSource(vertices, colors=colors)
    >>> sc.add_source(points)

    >>> im = sc.render()

    """

    _image = None
    data_source = None

    def __init__(self, positions, colors=None, color_stride=1, radii=None):
        assert positions.ndim == 2 and positions.shape[1] == 3
        if colors is not None:
            assert colors.ndim == 2 and colors.shape[1] == 4
            assert colors.shape[0] == positions.shape[0]
        if not is_sequence(radii):
            if radii is not None:  # broadcast the value
                radii = radii * np.ones(positions.shape[0], dtype="int64")
            else:  # default radii to 0 pixels (i.e. point is 1 pixel wide)
                radii = np.zeros(positions.shape[0], dtype="int64")
        else:
            assert radii.ndim == 1
            assert radii.shape[0] == positions.shape[0]
        self.positions = positions
        # If colors aren't individually set, make black with full opacity
        if colors is None:
            colors = np.ones((len(positions), 4))
        self.colors = colors
        self.color_stride = color_stride
        self.radii = radii

    def _validate(self):
        pass

    def render(self, camera, zbuffer=None):
        """Renders an image using the provided camera

        Parameters
        ----------
        camera: :class:`yt.visualization.volume_rendering.camera.Camera`
            A volume rendering camera. Can be any type of camera.
        zbuffer: :class:`yt.visualization.volume_rendering.zbuffer_array.Zbuffer`
            A zbuffer array. This is used for opaque sources to determine the
            z position of the source relative to other sources. Only useful if
            you are manually calling render on multiple sources. Scene.render
            uses this internally.

        Returns
        -------
        A :class:`yt.data_objects.image_array.ImageArray` containing
        the rendered image.

        """
        vertices = self.positions
        if zbuffer is None:
            empty = camera.lens.new_image(camera)
            z = np.empty(empty.shape[:2], dtype="float64")
            empty[:] = 0.0
            z[:] = np.inf
            zbuffer = ZBuffer(empty, z)
        else:
            empty = zbuffer.rgba
            z = zbuffer.z

        # DRAW SOME POINTS
        camera.lens.setup_box_properties(camera)
        px, py, dz = camera.lens.project_to_plane(camera, vertices)

        zpoints(empty, z, px, py, dz, self.colors, self.radii, self.color_stride)

        self.zbuffer = zbuffer
        return zbuffer

    def __repr__(self):
        disp = "<Point Source>"
        return disp


class LineSource(OpaqueSource):
    r"""A render source for a sequence of opaque line segments.

    This class provides a mechanism for adding lines to a scene; these
    points will be opaque, and can also be colored.

    .. note::

        If adding a LineSource to your rendering causes the image to appear
        blank or fades a VolumeSource, try lowering the values specified in
        the alpha channel of the ``colors`` array.

    Parameters
    ----------
    positions: array_like of shape (N, 2, 3)
        The positions of the starting and stopping points for each line.
        For example,positions[0][0] and positions[0][1] would give the (x, y, z)
        coordinates of the beginning and end points of the first line,
        respectively. If specified with no units, assumed to be in code units.
    colors : array_like of shape (N, 4), optional
        The colors of the points, including an alpha channel, in floating
        point running from 0..1.  The four channels correspond to r, g, b, and
        alpha values. Note that they correspond to the line segment succeeding
        each point; this means that strictly speaking they need only be (N-1)
        in length.
    color_stride : int, optional
        The stride with which to access the colors when putting them on the
        scene.

    Examples
    --------

    This example creates a volume rendering and then adds some random lines
    to the image:

    >>> import yt
    >>> import numpy as np
    >>> from yt.visualization.volume_rendering.api import LineSource
    >>> from yt.units import kpc
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

    >>> im, sc = yt.volume_render(ds)

    >>> nlines = 4
    >>> vertices = np.random.random([nlines, 2, 3]) * 600 * kpc
    >>> colors = np.random.random([nlines, 4])
    >>> colors[:, 3] = 1.0

    >>> lines = LineSource(vertices, colors)
    >>> sc.add_source(lines)

    >>> im = sc.render()

    """

    _image = None
    data_source = None

    def __init__(self, positions, colors=None, color_stride=1):
        super().__init__()

        assert positions.ndim == 3
        assert positions.shape[1] == 2
        assert positions.shape[2] == 3
        if colors is not None:
            assert colors.ndim == 2
            assert colors.shape[1] == 4

        # convert the positions to the shape expected by zlines, below
        N = positions.shape[0]
        self.positions = positions.reshape((2 * N, 3))

        # If colors aren't individually set, make black with full opacity
        if colors is None:
            colors = np.ones((len(positions), 4))
        self.colors = colors
        self.color_stride = color_stride

    def _validate(self):
        pass

    def render(self, camera, zbuffer=None):
        """Renders an image using the provided camera

        Parameters
        ----------
        camera: :class:`yt.visualization.volume_rendering.camera.Camera`
            A volume rendering camera. Can be any type of camera.
        zbuffer: :class:`yt.visualization.volume_rendering.zbuffer_array.Zbuffer`
            z position of the source relative to other sources. Only useful if
            you are manually calling render on multiple sources. Scene.render
            uses this internally.

        Returns
        -------
        A :class:`yt.data_objects.image_array.ImageArray` containing
        the rendered image.

        """
        vertices = self.positions
        if zbuffer is None:
            empty = camera.lens.new_image(camera)
            z = np.empty(empty.shape[:2], dtype="float64")
            empty[:] = 0.0
            z[:] = np.inf
            zbuffer = ZBuffer(empty, z)
        else:
            empty = zbuffer.rgba
            z = zbuffer.z

        # DRAW SOME LINES
        camera.lens.setup_box_properties(camera)
        px, py, dz = camera.lens.project_to_plane(camera, vertices)

        px = px.astype("int64")
        py = py.astype("int64")

        if len(px.shape) == 1:
            zlines(
                empty, z, px, py, dz, self.colors.astype("float64"), self.color_stride
            )
        else:
            # For stereo-lens, two sets of pos for each eye are contained
            # in px...pz
            zlines(
                empty,
                z,
                px[0, :],
                py[0, :],
                dz[0, :],
                self.colors.astype("float64"),
                self.color_stride,
            )
            zlines(
                empty,
                z,
                px[1, :],
                py[1, :],
                dz[1, :],
                self.colors.astype("float64"),
                self.color_stride,
            )

        self.zbuffer = zbuffer
        return zbuffer

    def __repr__(self):
        disp = "<Line Source>"
        return disp


class BoxSource(LineSource):
    r"""A render source for a box drawn with line segments.
    This render source will draw a box, with transparent faces, in data
    space coordinates.  This is useful for annotations.

    Parameters
    ----------
    left_edge: array-like of shape (3,), float
        The left edge coordinates of the box.
    right_edge : array-like of shape (3,), float
        The right edge coordinates of the box.
    color : array-like of shape (4,), float, optional
        The colors (including alpha) to use for the lines.
        Default is black with an alpha of 1.0.

    Examples
    --------

    This example shows how to use BoxSource to add an outline of the
    domain boundaries to a volume rendering.

    >>> import yt
    >>> from yt.visualization.volume_rendering.api import BoxSource
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

    >>> im, sc = yt.volume_render(ds)

    >>> box_source = BoxSource(
    ...     ds.domain_left_edge, ds.domain_right_edge, [1.0, 1.0, 1.0, 1.0]
    ... )
    >>> sc.add_source(box_source)

    >>> im = sc.render()

    """

    def __init__(self, left_edge, right_edge, color=None):
        assert left_edge.shape == (3,)
        assert right_edge.shape == (3,)

        if color is None:
            color = np.array([1.0, 1.0, 1.0, 1.0])

        color = ensure_numpy_array(color)
        color.shape = (1, 4)
        corners = get_corners(left_edge.copy(), right_edge.copy())
        order = [0, 1, 1, 2, 2, 3, 3, 0]
        order += [4, 5, 5, 6, 6, 7, 7, 4]
        order += [0, 4, 1, 5, 2, 6, 3, 7]
        vertices = np.empty([24, 3])
        for i in range(3):
            vertices[:, i] = corners[order, i, ...].ravel(order="F")
        vertices = vertices.reshape((12, 2, 3))

        super().__init__(vertices, color, color_stride=24)

    def _validate(self):
        pass


class GridSource(LineSource):
    r"""A render source for drawing grids in a scene.

    This render source will draw blocks that are within a given data
    source, by default coloring them by their level of resolution.

    Parameters
    ----------
    data_source: :class:`~yt.data_objects.api.DataContainer`
        The data container that will be used to identify grids to draw.
    alpha : float
        The opacity of the grids to draw.
    cmap : color map name
        The color map to use to map resolution levels to color.
    min_level : int, optional
        Minimum level to draw
    max_level : int, optional
        Maximum level to draw

    Examples
    --------

    This example makes a volume rendering and adds outlines of all the
    AMR grids in the simulation:

    >>> import yt
    >>> from yt.visualization.volume_rendering.api import GridSource
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

    >>> im, sc = yt.volume_render(ds)

    >>> grid_source = GridSource(ds.all_data(), alpha=1.0)

    >>> sc.add_source(grid_source)

    >>> im = sc.render()

    This example does the same thing, except it only draws the grids
    that are inside a sphere of radius (0.1, "unitary") located at the
    domain center:

    >>> import yt
    >>> from yt.visualization.volume_rendering.api import GridSource
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

    >>> im, sc = yt.volume_render(ds)

    >>> dd = ds.sphere("c", (0.1, "unitary"))
    >>> grid_source = GridSource(dd, alpha=1.0)

    >>> sc.add_source(grid_source)

    >>> im = sc.render()

    """

    def __init__(
        self, data_source, alpha=0.3, cmap=None, min_level=None, max_level=None
    ):
        self.data_source = data_source_or_all(data_source)
        corners = []
        levels = []
        for block, _mask in self.data_source.blocks:
            block_corners = np.array(
                [
                    [block.LeftEdge[0], block.LeftEdge[1], block.LeftEdge[2]],
                    [block.RightEdge[0], block.LeftEdge[1], block.LeftEdge[2]],
                    [block.RightEdge[0], block.RightEdge[1], block.LeftEdge[2]],
                    [block.LeftEdge[0], block.RightEdge[1], block.LeftEdge[2]],
                    [block.LeftEdge[0], block.LeftEdge[1], block.RightEdge[2]],
                    [block.RightEdge[0], block.LeftEdge[1], block.RightEdge[2]],
                    [block.RightEdge[0], block.RightEdge[1], block.RightEdge[2]],
                    [block.LeftEdge[0], block.RightEdge[1], block.RightEdge[2]],
                ],
                dtype="float64",
            )
            corners.append(block_corners)
            levels.append(block.Level)
        corners = np.dstack(corners)
        levels = np.array(levels)
        if cmap is None:
            cmap = ytcfg.get("yt", "default_colormap")

        if max_level is not None:
            subset = levels <= max_level
            levels = levels[subset]
            corners = corners[:, :, subset]
        if min_level is not None:
            subset = levels >= min_level
            levels = levels[subset]
            corners = corners[:, :, subset]

        colors = (
            apply_colormap(
                levels * 1.0,
                color_bounds=[0, self.data_source.ds.index.max_level],
                cmap_name=cmap,
            )[0, :, :]
            / 255.0
        )
        colors[:, 3] = alpha

        order = [0, 1, 1, 2, 2, 3, 3, 0]
        order += [4, 5, 5, 6, 6, 7, 7, 4]
        order += [0, 4, 1, 5, 2, 6, 3, 7]

        vertices = np.empty([corners.shape[2] * 2 * 12, 3])
        for i in range(3):
            vertices[:, i] = corners[order, i, ...].ravel(order="F")
        vertices = vertices.reshape((corners.shape[2] * 12, 2, 3))

        super().__init__(vertices, colors, color_stride=24)


class CoordinateVectorSource(OpaqueSource):
    r"""Draw coordinate vectors on the scene.

    This will draw a set of coordinate vectors on the camera image.  They
    will appear in the lower right of the image.

    Parameters
    ----------
    colors: array-like of shape (3,4), optional
        The RGBA values to use to draw the x, y, and z vectors. The default is
        [[1, 0, 0, alpha], [0, 1, 0, alpha], [0, 0, 1, alpha]]  where ``alpha``
        is set by the parameter below. If ``colors`` is set then ``alpha`` is
        ignored.
    alpha : float, optional
        The opacity of the vectors.
    thickness : int, optional
        The line thickness

    Examples
    --------

    >>> import yt
    >>> from yt.visualization.volume_rendering.api import \
    ...     CoordinateVectorSource
    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

    >>> im, sc = yt.volume_render(ds)

    >>> coord_source = CoordinateVectorSource()

    >>> sc.add_source(coord_source)

    >>> im = sc.render()

    """

    def __init__(self, colors=None, alpha=1.0, *, thickness=1):
        super().__init__()
        # If colors aren't individually set, make black with full opacity
        if colors is None:
            colors = np.zeros((3, 4))
            colors[0, 0] = 1.0  # x is red
            colors[1, 1] = 1.0  # y is green
            colors[2, 2] = 1.0  # z is blue
            colors[:, 3] = alpha
        self.colors = colors
        self.thick = thickness

    def _validate(self):
        pass

    def render(self, camera, zbuffer=None):
        """Renders an image using the provided camera

        Parameters
        ----------
        camera: :class:`yt.visualization.volume_rendering.camera.Camera`
            A volume rendering camera. Can be any type of camera.
        zbuffer: :class:`yt.visualization.volume_rendering.zbuffer_array.Zbuffer`
            A zbuffer array. This is used for opaque sources to determine the
            z position of the source relative to other sources. Only useful if
            you are manually calling render on multiple sources. Scene.render
            uses this internally.

        Returns
        -------
        A :class:`yt.data_objects.image_array.ImageArray` containing
        the rendered image.

        """
        camera.lens.setup_box_properties(camera)
        center = camera.focus
        # Get positions at the focus
        positions = np.zeros([6, 3])
        positions[:] = center

        # Create vectors in the x,y,z directions
        for i in range(3):
            positions[2 * i + 1, i] += camera.width.in_units("code_length").d[i] / 16.0

        # Project to the image plane
        px, py, dz = camera.lens.project_to_plane(camera, positions)

        if len(px.shape) == 1:
            dpx = px[1::2] - px[::2]
            dpy = py[1::2] - py[::2]

            # Set the center of the coordinates to be in the lower left of the image
            lpx = camera.resolution[0] / 8
            lpy = camera.resolution[1] - camera.resolution[1] / 8  # Upside-downsies

            # Offset the pixels according to the projections above
            px[::2] = lpx
            px[1::2] = lpx + dpx
            py[::2] = lpy
            py[1::2] = lpy + dpy
            dz[:] = 0.0
        else:
            # For stereo-lens, two sets of pos for each eye are contained in px...pz
            dpx = px[:, 1::2] - px[:, ::2]
            dpy = py[:, 1::2] - py[:, ::2]

            lpx = camera.resolution[0] / 16
            lpy = camera.resolution[1] - camera.resolution[1] / 8  # Upside-downsies

            # Offset the pixels according to the projections above
            px[:, ::2] = lpx
            px[:, 1::2] = lpx + dpx
            px[1, :] += camera.resolution[0] / 2
            py[:, ::2] = lpy
            py[:, 1::2] = lpy + dpy
            dz[:, :] = 0.0

        # Create a zbuffer if needed
        if zbuffer is None:
            empty = camera.lens.new_image(camera)
            z = np.empty(empty.shape[:2], dtype="float64")
            empty[:] = 0.0
            z[:] = np.inf
            zbuffer = ZBuffer(empty, z)
        else:
            empty = zbuffer.rgba
            z = zbuffer.z

        # Draw the vectors

        px = px.astype("int64")
        py = py.astype("int64")

        if len(px.shape) == 1:
            zlines(
                empty, z, px, py, dz, self.colors.astype("float64"), thick=self.thick
            )
        else:
            # For stereo-lens, two sets of pos for each eye are contained
            # in px...pz
            zlines(
                empty,
                z,
                px[0, :],
                py[0, :],
                dz[0, :],
                self.colors.astype("float64"),
                thick=self.thick,
            )
            zlines(
                empty,
                z,
                px[1, :],
                py[1, :],
                dz[1, :],
                self.colors.astype("float64"),
                thick=self.thick,
            )

        # Set the new zbuffer
        self.zbuffer = zbuffer
        return zbuffer

    def __repr__(self):
        disp = "<Coordinates Source>"
        return disp



import abc
import glob
import os
from functools import cached_property

from yt.config import ytcfg
from yt.funcs import mylog
from yt.utilities.cython_fortran_utils import FortranFile

from .io import _read_fluid_file_descriptor
from .io_utils import read_offset

FIELD_HANDLERS: set[type["FieldFileHandler"]] = set()


def get_field_handlers():
    return FIELD_HANDLERS


def register_field_handler(ph):
    FIELD_HANDLERS.add(ph)


DETECTED_FIELDS = {}  # type: ignore


class HandlerMixin:
    """This contains all the shared methods to handle RAMSES files.

    This is not supposed to be user-facing.
    """

    def setup_handler(self, domain):
        """
        Initialize an instance of the class. This automatically sets
        the full path to the file. This is not intended to be
        overridden in most cases.

        If you need more flexibility, rewrite this function to your
        need in the inherited class.
        """
        self.ds = ds = domain.ds
        self.domain = domain
        self.domain_id = domain.domain_id

        basename = os.path.abspath(ds.root_folder)
        iout = int(os.path.basename(ds.parameter_filename).split(".")[0].split("_")[1])

        if ds.num_groups > 0:
            igroup = ((domain.domain_id - 1) // ds.group_size) + 1
            full_path = os.path.join(
                basename,
                f"group_{igroup:05d}",
                self.fname.format(iout=iout, icpu=domain.domain_id),
            )
        else:
            full_path = os.path.join(
                basename, self.fname.format(iout=iout, icpu=domain.domain_id)
            )

        if os.path.exists(full_path):
            self.fname = full_path
        else:
            raise FileNotFoundError(
                f"Could not find {self._file_type} file (type: {self.ftype}). "
                f"Tried {full_path}"
            )

        if self.file_descriptor is not None:
            if ds.num_groups > 0:
                # The particle file descriptor is *only* in the first group
                self.file_descriptor = os.path.join(
                    basename, "group_00001", self.file_descriptor
                )
            else:
                self.file_descriptor = os.path.join(basename, self.file_descriptor)

    @property
    def exists(self):
        """
        This function should return True if the *file* the instance
        exists. It is called for each file of the type found on the
        disk.

        By default, it just returns whether the file exists. Override
        it for more complex cases.
        """
        return os.path.exists(self.fname)

    @property
    def has_descriptor(self):
        """
        This function should return True if a *file descriptor*
        exists.

        By default, it just returns whether the file exists. Override
        it for more complex cases.
        """
        return os.path.exists(self.file_descriptor)

    @classmethod
    def any_exist(cls, ds):
        """
        This function should return True if the kind of particle
        represented by the class exists in the dataset. It takes as
        argument the class itself -not an instance- and a dataset.

        Arguments
        ---------
        ds : a Ramses Dataset

        Note
        ----
        This function is usually called once at the initialization of
        the RAMSES Dataset structure to determine if the particle type
        (e.g. regular particles) exists.
        """
        if ds.unique_identifier in cls._unique_registry:
            return cls._unique_registry[ds.unique_identifier]

        iout = int(os.path.basename(ds.parameter_filename).split(".")[0].split("_")[1])

        fname = os.path.join(
            os.path.split(ds.parameter_filename)[0], cls.fname.format(iout=iout, icpu=1)
        )
        exists = os.path.exists(fname)
        cls._unique_registry[ds.unique_identifier] = exists

        return exists


class FieldFileHandler(abc.ABC, HandlerMixin):
    """
    Abstract class to handle particles in RAMSES. Each instance
    represents a single file (one domain).

    To add support to a new particle file, inherit from this class and
    implement all functions containing a `NotImplementedError`.

    See `SinkParticleFileHandler` for an example implementation."""

    _file_type = "field"

    # These properties are static properties
    ftype: str | None = None  # The name to give to the field type
    fname: str | None = None  # The name of the file(s)
    # The attributes of the header
    attrs: tuple[tuple[str, int, str], ...] | None = None
    known_fields = None  # A list of tuple containing the field name and its type
    config_field: str | None = None  # Name of the config section (if any)

    file_descriptor: str | None = None  # The name of the file descriptor (if any)

    # These properties are computed dynamically
    field_offsets = None  # Mapping from field to offset in file
    field_types = (
        None  # Mapping from field to the type of the data (float, integer, ...)
    )

    def __init_subclass__(cls, *args, **kwargs):
        """
        Registers subclasses at creation.
        """
        super().__init_subclass__(*args, **kwargs)

        if cls.ftype is not None:
            register_field_handler(cls)

        cls._unique_registry = {}
        cls.parameters = {}
        cls.rt_parameters = {}
        cls._detected_field_list = {}
        return cls

    def __init__(self, domain):
        self.setup_handler(domain)

    @classmethod
    @abc.abstractmethod
    def detect_fields(cls, ds):
        """
        Called once to setup the fields of this type

        It should set the following static variables:

        * parameters: dictionary
            Dictionary containing the variables. The keys should match
            those of `cls.attrs`
        * field_list: list of (ftype, fname)
            The list of the field present in the file
        """
        pass

    @classmethod
    def get_detected_fields(cls, ds):
        """
        Get the detected fields from the registry.
        """
        if ds.unique_identifier in DETECTED_FIELDS:
            d = DETECTED_FIELDS[ds.unique_identifier]
            if cls.ftype in d:
                return d[cls.ftype]

        return None

    @classmethod
    def set_detected_fields(cls, ds, fields):
        """
        Store the detected fields into the registry.
        """
        if ds.unique_identifier not in DETECTED_FIELDS:
            DETECTED_FIELDS[ds.unique_identifier] = {}

        DETECTED_FIELDS[ds.unique_identifier].update({cls.ftype: fields})

    @classmethod
    def purge_detected_fields(cls, ds):
        """
        Purge the registry.

        This should be called on dataset creation to force the field
        detection to be called.
        """
        if ds.unique_identifier in DETECTED_FIELDS:
            DETECTED_FIELDS.pop(ds.unique_identifier)

    @property
    def level_count(self):
        """
        Return the number of cells per level.
        """
        if getattr(self, "_level_count", None) is not None:
            return self._level_count
        self.offset

        return self._level_count

    @property
    def field_list(self):
        return self._detected_field_list[self.ds.unique_identifier]

    @cached_property
    def offset(self):
        """
        Compute the offsets of the fields.

        By default, it skips the header (as defined by `cls.attrs`)
        and computes the offset at each level.

        It should be generic enough for most of the cases, but if the
        *structure* of your fluid file is non-canonical, change this.
        """
        nvars = len(self._detected_field_list[self.ds.unique_identifier])
        with FortranFile(self.fname) as fd:
            # Skip headers
            nskip = len(self.attrs)
            fd.skip(nskip)
            min_level = self.domain.ds.min_level

            # The file is as follows:
            # > headers
            # loop over levels
            #   loop over cpu domains
            #     > <ilevel>: current level
            #     > <nocts>: number of octs in level, domain
            #     loop over <nvars> variables (positions, velocities, density, ...)
            #       loop over <2*2*2> cells in each oct
            #          > <data> with shape (nocts, )
            #
            # So there are 8 * nvars records each with length (nocts, )
            # at each (level, cpus)
            offset, level_count = read_offset(
                fd,
                min_level,
                self.domain.domain_id,
                self.parameters[self.ds.unique_identifier]["nvar"],
                self.domain.amr_header,
                Nskip=nvars * 8,
            )

        self._level_count = level_count
        return offset

    @classmethod
    def load_fields_from_yt_config(cls) -> list[str]:
        if cls.config_field and ytcfg.has_section(cls.config_field):
            cfg = ytcfg.get(cls.config_field, "fields")
            fields = [_.strip() for _ in cfg if _.strip() != ""]
            return fields

        return []


class HydroFieldFileHandler(FieldFileHandler):
    ftype = "ramses"
    fname = "hydro_{iout:05d}.out{icpu:05d}"
    file_descriptor = "hydro_file_descriptor.txt"
    config_field = "ramses-hydro"

    attrs = (
        ("ncpu", 1, "i"),
        ("nvar", 1, "i"),
        ("ndim", 1, "i"),
        ("nlevelmax", 1, "i"),
        ("nboundary", 1, "i"),
        ("gamma", 1, "d"),
    )

    @classmethod
    def detect_fields(cls, ds):
        # Try to get the detected fields
        detected_fields = cls.get_detected_fields(ds)
        if detected_fields:
            return detected_fields

        num = os.path.basename(ds.parameter_filename).split(".")[0].split("_")[1]
        testdomain = 1  # Just pick the first domain file to read
        basepath = os.path.abspath(os.path.dirname(ds.parameter_filename))
        basename = "%s/%%s_%s.out%05i" % (basepath, num, testdomain)
        fname = basename % "hydro"
        fname_desc = os.path.join(basepath, cls.file_descriptor)

        attrs = cls.attrs
        with FortranFile(fname) as fd:
            hvals = fd.read_attrs(attrs)
        cls.parameters[ds.unique_identifier] = hvals

        # Store some metadata
        ds.gamma = hvals["gamma"]
        nvar = hvals["nvar"]

        ok = False

        if ds._fields_in_file is not None:
            # Case 1: fields are provided by users on construction of dataset
            fields = list(ds._fields_in_file)
            ok = True
        else:
            # Case 2: fields are provided by users in the config
            fields = cls.load_fields_from_yt_config()
            ok = len(fields) > 0

        if not ok and os.path.exists(fname_desc):
            # Case 3: there is a file descriptor
            # Or there is an hydro file descriptor
            mylog.debug("Reading hydro file descriptor.")
            # For now, we can only read double precision fields
            fields = [
                e[0] for e in _read_fluid_file_descriptor(fname_desc, prefix="hydro")
            ]

            # We get no fields for old-style hydro file descriptor
            ok = len(fields) > 0

        if not ok:
            # Case 4: attempt autodetection with usual fields
            foldername = os.path.abspath(os.path.dirname(ds.parameter_filename))
            rt_flag = any(glob.glob(os.sep.join([foldername, "info_rt_*.txt"])))
            if rt_flag:  # rt run
                if nvar < 10:
                    mylog.info("Detected RAMSES-RT file WITHOUT IR trapping.")

                    fields = [
                        "Density",
                        "x-velocity",
                        "y-velocity",
                        "z-velocity",
                        "Pressure",
                        "Metallicity",
                        "HII",
                        "HeII",
                        "HeIII",
                    ]
                else:
                    mylog.info("Detected RAMSES-RT file WITH IR trapping.")

                    fields = [
                        "Density",
                        "x-velocity",
                        "y-velocity",
                        "z-velocity",
                        "Pres_IR",
                        "Pressure",
                        "Metallicity",
                        "HII",
                        "HeII",
                        "HeIII",
                    ]
            else:
                if nvar < 5:
                    mylog.debug(
                        "nvar=%s is too small! YT doesn't currently "
                        "support 1D/2D runs in RAMSES %s"
                    )
                    raise ValueError
                # Basic hydro runs
                if nvar == 5:
                    fields = [
                        "Density",
                        "x-velocity",
                        "y-velocity",
                        "z-velocity",
                        "Pressure",
                    ]
                if nvar > 5 and nvar < 11:
                    fields = [
                        "Density",
                        "x-velocity",
                        "y-velocity",
                        "z-velocity",
                        "Pressure",
                        "Metallicity",
                    ]
                # MHD runs - NOTE:
                # THE MHD MODULE WILL SILENTLY ADD 3 TO THE NVAR IN THE MAKEFILE
                if nvar == 11:
                    fields = [
                        "Density",
                        "x-velocity",
                        "y-velocity",
                        "z-velocity",
                        "B_x_left",
                        "B_y_left",
                        "B_z_left",
                        "B_x_right",
                        "B_y_right",
                        "B_z_right",
                        "Pressure",
                    ]
                if nvar > 11:
                    fields = [
                        "Density",
                        "x-velocity",
                        "y-velocity",
                        "z-velocity",
                        "B_x_left",
                        "B_y_left",
                        "B_z_left",
                        "B_x_right",
                        "B_y_right",
                        "B_z_right",
                        "Pressure",
                        "Metallicity",
                    ]
            mylog.debug(
                "No fields specified by user; automatically setting fields array to %s",
                fields,
            )

        # Allow some wiggle room for users to add too many variables
        count_extra = 0
        while len(fields) < nvar:
            fields.append(f"var_{len(fields)}")
            count_extra += 1
        if count_extra > 0:
            mylog.debug("Detected %s extra fluid fields.", count_extra)
        cls._detected_field_list[ds.unique_identifier] = [
            (cls.ftype, e) for e in fields
        ]

        cls.set_detected_fields(ds, fields)

        return fields


class GravFieldFileHandler(FieldFileHandler):
    ftype = "gravity"
    fname = "grav_{iout:05d}.out{icpu:05d}"
    config_field = "ramses-grav"

    attrs = (
        ("ncpu", 1, "i"),
        ("nvar", 1, "i"),
        ("nlevelmax", 1, "i"),
        ("nboundary", 1, "i"),
    )

    @classmethod
    def detect_fields(cls, ds):
        # Try to get the detected fields
        detected_fields = cls.get_detected_fields(ds)
        if detected_fields:
            return detected_fields

        ndim = ds.dimensionality
        iout = int(str(ds).split("_")[1])
        basedir = os.path.split(ds.parameter_filename)[0]
        fname = os.path.join(basedir, cls.fname.format(iout=iout, icpu=1))
        with FortranFile(fname) as fd:
            cls.parameters[ds.unique_identifier] = fd.read_attrs(cls.attrs)

        nvar = cls.parameters[ds.unique_identifier]["nvar"]
        ndim = ds.dimensionality

        fields = cls.load_fields_from_yt_config()

        if not fields:
            if nvar == ndim + 1:
                fields = ["Potential"] + [f"{k}-acceleration" for k in "xyz"[:ndim]]
            else:
                fields = [f"{k}-acceleration" for k in "xyz"[:ndim]]
            ndetected = len(fields)

            if ndetected != nvar and not ds._warned_extra_fields["gravity"]:
                mylog.info("Detected %s extra gravity fields.", nvar - ndetected)
                ds._warned_extra_fields["gravity"] = True

                for i in range(nvar - ndetected):
                    fields.append(f"var{i}")

        cls._detected_field_list[ds.unique_identifier] = [
            (cls.ftype, e) for e in fields
        ]

        cls.set_detected_fields(ds, fields)

        return fields


class RTFieldFileHandler(FieldFileHandler):
    ftype = "ramses-rt"
    fname = "rt_{iout:05d}.out{icpu:05d}"
    file_descriptor = "rt_file_descriptor.txt"
    config_field = "ramses-rt"

    attrs = (
        ("ncpu", 1, "i"),
        ("nvar", 1, "i"),
        ("ndim", 1, "i"),
        ("nlevelmax", 1, "i"),
        ("nboundary", 1, "i"),
        ("gamma", 1, "d"),
    )

    @classmethod
    def detect_fields(cls, ds):
        # Try to get the detected fields
        detected_fields = cls.get_detected_fields(ds)
        if detected_fields:
            return detected_fields

        fname = ds.parameter_filename.replace("info_", "info_rt_")

        rheader = {}

        def read_rhs(cast):
            line = f.readline()
            p, v = line.split("=")
            rheader[p.strip()] = cast(v)

        with open(fname) as f:
            # Read nRTvar, nions, ngroups, iions
            for _ in range(4):
                read_rhs(int)

            # Try to read rtprecision.
            # Either it is present or the line is simply blank, so
            # we try to parse the line as an int, and if it fails,
            # we simply ignore it.
            try:
                read_rhs(int)
                f.readline()
            except ValueError:
                pass

            # Read X and Y fractions
            for _ in range(2):
                read_rhs(float)
            f.readline()

            # Reat unit_np, unit_pfd
            for _ in range(2):
                read_rhs(float)

            # Read rt_c_frac
            # Note: when using variable speed of light, this line will contain multiple
            # values corresponding the the velocity at each level
            read_rhs(lambda line: [float(e) for e in line.split()])
            f.readline()

            # Read n star, t2star, g_star
            for _ in range(3):
                read_rhs(float)

            # Touchy part, we have to read the photon group properties
            mylog.debug("Not reading photon group properties")

            cls.rt_parameters[ds.unique_identifier] = rheader

        ngroups = rheader["nGroups"]

        iout = int(str(ds).split("_")[1])
        basedir = os.path.split(ds.parameter_filename)[0]
        fname = os.path.join(basedir, cls.fname.format(iout=iout, icpu=1))
        fname_desc = os.path.join(basedir, cls.file_descriptor)
        with FortranFile(fname) as fd:
            cls.parameters[ds.unique_identifier] = fd.read_attrs(cls.attrs)

        ok = False

        if ds._fields_in_file is not None:
            # Case 1: fields are provided by users on construction of dataset
            fields = list(ds._fields_in_file)
            ok = True
        else:
            # Case 2: fields are provided by users in the config
            fields = cls.load_fields_from_yt_config()
            ok = len(fields) > 0

        if not ok and os.path.exists(fname_desc):
            # Case 3: there is a file descriptor
            # Or there is an hydro file descriptor
            mylog.debug("Reading rt file descriptor.")
            # For now, we can only read double precision fields
            fields = [
                e[0] for e in _read_fluid_file_descriptor(fname_desc, prefix="rt")
            ]
            ok = len(fields) > 0

        if not ok:
            fields = []
            tmp = [
                "Photon_density_%s",
                "Photon_flux_x_%s",
                "Photon_flux_y_%s",
                "Photon_flux_z_%s",
            ]
            for ng in range(ngroups):
                fields.extend([t % (ng + 1) for t in tmp])

        cls._detected_field_list[ds.unique_identifier] = [
            (cls.ftype, e) for e in fields
        ]

        cls.set_detected_fields(ds, fields)
        return fields

    @classmethod
    def get_rt_parameters(cls, ds):
        if cls.rt_parameters[ds.unique_identifier]:
            return cls.rt_parameters[ds.unique_identifier]

        # Call detect fields to get the rt_parameters
        cls.detect_fields(ds)
        return cls.rt_parameters[ds.unique_identifier]



import numpy as np
import pytest

from yt import SlicePlot, load_uniform_grid
from yt.testing import fake_amr_ds, requires_module


@requires_module("cartopy")
@pytest.mark.parametrize("geometry", ["geographic", "internal_geographic"])
def test_quiver_callback_geographic(geometry):
    flds = ("density", "velocity_ew", "velocity_ns")
    units = ("g/cm**3", "m/s", "m/s")

    ds = fake_amr_ds(fields=flds, units=units, geometry=geometry)

    for ax in ds.coordinates.axis_order:
        slc = SlicePlot(ds, ax, "density", buff_size=(50, 50))
        if ax == ds.coordinates.radial_axis:
            # avoid the exact transform bounds
            slc.set_width((359.99, 179.99))
        slc.annotate_quiver(("stream", "velocity_ew"), ("stream", "velocity_ns"))
        slc.render()


@pytest.fixture()
def ds_geo_uni_grid():
    yc = 0.0
    xc = 0.0

    def _vel_calculator(grid, ax):
        y_lat = grid.fcoords[:, 1].d
        x_lon = grid.fcoords[:, 2].d
        x_lon[x_lon > 180] = x_lon[x_lon > 180] - 360.0
        dist = np.sqrt((y_lat - yc) ** 2 + (xc - x_lon) ** 2)
        if ax == 1:
            sgn = np.sign(y_lat - yc)
        elif ax == 2:
            sgn = np.sign(x_lon - xc)

        vel = np.exp(-((dist / 45) ** 2)) * sgn
        return vel.reshape(grid.shape)

    def _calculate_u(grid, field):
        return _vel_calculator(grid, 2)

    def _calculate_v(grid, field):
        return _vel_calculator(grid, 1)

    data = {"u_vel": _calculate_u, "v_vel": _calculate_v}

    bbox = [[10.0, 1000], [-90, 90], [0, 360]]
    bbox = np.array(bbox)

    ds = load_uniform_grid(
        data,
        (16, 16, 32),
        bbox=bbox,
        geometry="geographic",
        axis_order=("altitude", "latitude", "longitude"),
    )

    return ds


@requires_module("cartopy")
@pytest.mark.mpl_image_compare
def test_geoquiver_answer(ds_geo_uni_grid):
    slc = SlicePlot(ds_geo_uni_grid, "altitude", "u_vel")
    slc.set_width((359.99, 179.99))
    slc.set_log("u_vel", False)
    slc.annotate_quiver("u_vel", "v_vel", scale=50)
    slc.render()
    return slc.plots["u_vel"].figure



"""
A proxy object for field descriptors, usually living as ds.fields.
"""

import inspect
import textwrap
import weakref
from functools import cached_property

from yt._maintenance.ipython_compat import IPYWIDGETS_ENABLED
from yt.fields.derived_field import DerivedField


def _fill_values(values):
    value = (
        '<div class="rendered_html jp-RenderedHTMLCommon">'
        + "<table><thead><tr><th>Name</th><th>Type</th>"
        + "<th>Value</th></tr></thead><tr><td>"
        + "</td></tr><tr><td>".join(
            f"{v}</td><td>{type(values[v]).__name__}</td><td>{str(values[v])}"
            for v in sorted(values)
        )
        + "</td></tr></table></div>"
    )
    return value


class FieldTypeContainer:
    def __init__(self, ds):
        self.ds = weakref.proxy(ds)

    def __getattr__(self, attr):
        ds = self.__getattribute__("ds")
        fnc = FieldNameContainer(ds, attr)
        if len(dir(fnc)) == 0:
            return self.__getattribute__(attr)
        return fnc

    @cached_property
    def field_types(self):
        return {t for t, n in self.ds.field_info}

    def __dir__(self):
        return list(self.field_types)

    def __iter__(self):
        for ft in self.field_types:
            fnc = FieldNameContainer(self.ds, ft)
            if len(dir(fnc)) == 0:
                yield self.__getattribute__(ft)
            else:
                yield fnc

    def __contains__(self, obj):
        ob = None
        if isinstance(obj, FieldNameContainer):
            ob = obj.field_type
        elif isinstance(obj, str):
            ob = obj

        return ob in self.field_types

    if IPYWIDGETS_ENABLED:

        def _ipython_display_(self):
            import ipywidgets
            from IPython.display import display

            fnames = []
            children = []
            for ftype in sorted(self.field_types):
                fnc = getattr(self, ftype)
                children.append(ipywidgets.Output())
                with children[-1]:
                    display(fnc)
                fnames.append(ftype)
            tabs = ipywidgets.Tab(children=children)
            for i, n in enumerate(fnames):
                tabs.set_title(i, n)
            display(tabs)


class FieldNameContainer:
    def __init__(self, ds, field_type):
        self.ds = ds
        self.field_type = field_type

    def __getattr__(self, attr):
        ft = self.__getattribute__("field_type")
        ds = self.__getattribute__("ds")
        if (ft, attr) not in ds.field_info:
            return self.__getattribute__(attr)
        return ds.field_info[ft, attr]

    def __dir__(self):
        return [n for t, n in self.ds.field_info if t == self.field_type]

    def __iter__(self):
        for t, n in self.ds.field_info:
            if t == self.field_type:
                yield self.ds.field_info[t, n]

    def __contains__(self, obj):
        if isinstance(obj, DerivedField):
            if self.field_type == obj.name[0] and obj.name in self.ds.field_info:
                # e.g. from a completely different dataset
                if self.ds.field_info[obj.name] is not obj:
                    return False
                return True
        elif isinstance(obj, tuple):
            if self.field_type == obj[0] and obj in self.ds.field_info:
                return True
        elif isinstance(obj, str):
            if (self.field_type, obj) in self.ds.field_info:
                return True
        return False

    if IPYWIDGETS_ENABLED:
        # for discussion of this class-level conditional: https://github.com/yt-project/yt/pull/4941

        def _ipython_display_(self):
            import ipywidgets
            from IPython.display import Markdown, display

            names = dir(self)
            names.sort()

            def change_field(_ftype, _box, _var_window):
                def _change_field(event):
                    fobj = getattr(_ftype, event["new"])
                    _box.clear_output()
                    with _box:
                        display(
                            Markdown(
                                data="```python\n"
                                + textwrap.dedent(fobj.get_source())
                                + "\n```"
                            )
                        )
                    values = inspect.getclosurevars(fobj._function).nonlocals
                    _var_window.value = _fill_values(values)

                return _change_field

            flist = ipywidgets.Select(
                options=names, layout=ipywidgets.Layout(height="95%")
            )
            source = ipywidgets.Output(
                layout=ipywidgets.Layout(width="100%", height="9em")
            )
            var_window = ipywidgets.HTML(value="Empty")
            var_box = ipywidgets.Box(
                layout=ipywidgets.Layout(
                    width="100%", height="100%", overflow_y="scroll"
                )
            )
            var_box.children = [var_window]
            ftype_tabs = ipywidgets.Tab(
                children=[source, var_box],
                layout=ipywidgets.Layout(flex="2 1 auto", width="auto", height="95%"),
            )
            ftype_tabs.set_title(0, "Source")
            ftype_tabs.set_title(1, "Variables")
            flist.observe(change_field(self, source, var_window), "value")
            display(
                ipywidgets.HBox(
                    [flist, ftype_tabs], layout=ipywidgets.Layout(height="14em")
                )
            )



# distutils: libraries = STD_LIBS
# distutils: include_dirs = LIB_DIR
cimport cython
from libc.stdio cimport SEEK_CUR, SEEK_SET
cimport numpy as np

import numpy as np

from yt.geometry.oct_container cimport RAMSESOctreeContainer
from yt.utilities.cython_fortran_utils cimport FortranFile

from yt.utilities.exceptions import YTIllDefinedAMRData


ctypedef np.int32_t INT32_t
ctypedef np.int64_t INT64_t
ctypedef np.float64_t DOUBLE_t

cdef int INT32_SIZE = sizeof(np.int32_t)
cdef int INT64_SIZE = sizeof(np.int64_t)
cdef int DOUBLE_SIZE = sizeof(np.float64_t)


cdef inline int skip_len(int Nskip, int record_len) noexcept nogil:
    return Nskip * (record_len * DOUBLE_SIZE + INT64_SIZE)


@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def read_amr(FortranFile f, dict headers,
             np.ndarray[np.int64_t, ndim=1] ngridbound, INT64_t min_level,
             RAMSESOctreeContainer oct_handler):

    cdef INT64_t ncpu, nboundary, max_level, nlevelmax, ncpu_and_bound
    cdef DOUBLE_t nx, ny, nz
    cdef INT64_t ilevel, icpu, n, ndim, jump_len
    cdef INT32_t ng
    cdef np.ndarray[np.int32_t, ndim=2] numbl
    cdef np.ndarray[np.float64_t, ndim=2] pos
    cdef int i

    # The ordering is very important here, as we'll write directly into the memory
    # address the content of the files.
    cdef np.float64_t[::1, :] pos_view

    ndim = headers['ndim']
    numbl = headers['numbl']
    nboundary = headers['nboundary']
    nx, ny, nz = (((i-1.0)/2.0) for i in headers['nx'])
    nlevelmax = headers['nlevelmax']
    ncpu = headers['ncpu']

    ncpu_and_bound = nboundary + ncpu

    # Allocate more memory if required
    pos = np.empty((max(numbl.max(), ngridbound.max()), 3), dtype="d", order="F")
    pos_view = pos

    # Compute number of fields to skip. This should be 31 in 3 dimensions
    jump_len = (1          # father index
                + 2*ndim   # neighbor index
                + 2**ndim  # son index
                + 2**ndim  # cpu map
                + 2**ndim  # refinement map
    )
    # Initialize values
    max_level = 0
    cdef int record_len

    for ilevel in range(nlevelmax):
        for icpu in range(ncpu_and_bound):
            if icpu < ncpu:
                ng = numbl[ilevel, icpu]
            else:
                ng = ngridbound[icpu - ncpu + nboundary*ilevel]

            if ng == 0:
                continue
            # Skip grid index, 'next' and 'prev' arrays
            record_len = (ng * INT32_SIZE + INT64_SIZE)
            f.seek(record_len * 3, SEEK_CUR)

            f.read_vector_inplace("d", <void*> &pos_view[0, 0])
            f.read_vector_inplace("d", <void*> &pos_view[0, 1])
            f.read_vector_inplace("d", <void*> &pos_view[0, 2])

            for i in range(ng):
                pos_view[i, 0] -= nx
            for i in range(ng):
                pos_view[i, 1] -= ny
            for i in range(ng):
                pos_view[i, 2] -= nz

            # Skip father, neighbor, son, cpu map and refinement map
            f.seek(record_len * jump_len, SEEK_CUR)

            # Note that we're adding *grids*, not individual cells.
            if ilevel >= min_level:
                n = oct_handler.add(icpu + 1, ilevel - min_level, pos[:ng, :],
                                    count_boundary = 1)
                if n > 0:
                    max_level = max(ilevel - min_level, max_level)

    return max_level

@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cpdef read_offset(FortranFile f, INT64_t min_level, INT64_t domain_id, INT64_t nvar, dict headers, int Nskip):

    cdef np.ndarray[np.int64_t, ndim=2] offset, level_count
    cdef INT64_t ndim, twotondim, nlevelmax, n_levels, nboundary, ncpu, ncpu_and_bound
    cdef INT64_t ilevel, icpu
    cdef INT32_t file_ilevel, file_ncache

    ndim = headers['ndim']
    nboundary = headers['nboundary']
    nlevelmax = headers['nlevelmax']
    n_levels = nlevelmax - min_level
    ncpu = headers['ncpu']

    ncpu_and_bound = nboundary + ncpu
    twotondim = 2**ndim

    if Nskip == -1:
        Nskip = twotondim * nvar

    # It goes: level, CPU, 8-variable (1 oct)
    offset = np.full((ncpu_and_bound, n_levels), -1, dtype=np.int64)
    level_count = np.zeros((ncpu_and_bound, n_levels), dtype=np.int64)

    cdef np.int64_t[:, ::1] level_count_view = level_count
    cdef np.int64_t[:, ::1] offset_view = offset

    for ilevel in range(nlevelmax):
        for icpu in range(ncpu_and_bound):
            file_ilevel = f.read_int()
            file_ncache = f.read_int()
            if file_ncache == 0:
                continue

            if file_ilevel != ilevel+1:
                raise YTIllDefinedAMRData(
                    'Cannot read offsets in file %s. The level read '
                    'from data (%s) is not coherent with the expected (%s)',
                    f.name, file_ilevel, ilevel)

            if ilevel >= min_level:
                offset_view[icpu, ilevel - min_level] = f.tell()
                level_count_view[icpu, ilevel - min_level] = <INT64_t> file_ncache
            f.seek(skip_len(Nskip, file_ncache), SEEK_CUR)

    return offset, level_count

@cython.cpow(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def fill_hydro(FortranFile f,
               np.ndarray[np.int64_t, ndim=2] offsets,
               np.ndarray[np.int64_t, ndim=2] level_count,
               list cpu_enumerator,
               np.ndarray[np.uint8_t, ndim=1] level_inds,
               np.ndarray[np.uint8_t, ndim=1] cell_inds,
               np.ndarray[np.int64_t, ndim=1] file_inds,
               INT64_t ndim, list all_fields, list fields,
               dict tr,
               RAMSESOctreeContainer oct_handler,
               np.ndarray[np.int32_t, ndim=1] domain_inds=np.array([], dtype='int32')):
    cdef INT64_t offset
    cdef dict tmp
    cdef str field
    cdef INT64_t twotondim
    cdef int ilevel, icpu, nlevels, nc, ncpu_selected, nfields_selected
    cdef int i, j, ii

    twotondim = 2**ndim
    nfields_selected = len(fields)

    nlevels = offsets.shape[1]
    ncpu_selected = len(cpu_enumerator)

    cdef np.int64_t[::1] cpu_list = np.asarray(cpu_enumerator, dtype=np.int64)

    cdef np.int64_t[::1] jumps = np.zeros(nfields_selected + 1, dtype=np.int64)
    cdef int jump_len, Ncells
    cdef np.uint8_t[::1] mask_level = np.zeros(nlevels, dtype=np.uint8)

    # First, make sure fields are in the same order
    fields = sorted(fields, key=lambda f: all_fields.index(f))

    # The ordering is very important here, as we'll write directly into the memory
    # address the content of the files.
    cdef np.float64_t[::1, :, :] buffer

    jump_len = 0
    j = 0
    for i, field in enumerate(all_fields):
        if field in fields:
            jumps[j] = jump_len
            j += 1
            jump_len = 0
        else:
            jump_len += 1
    jumps[j] = jump_len
    cdef int first_field_index = jumps[0]

    buffer = np.empty((level_count.max(), twotondim, nfields_selected), dtype="float64", order='F')

    # Precompute which levels we need to read
    Ncells = len(level_inds)
    for i in range(Ncells):
        mask_level[level_inds[i]] |= 1

    # Loop over levels
    for ilevel in range(nlevels):
        if mask_level[ilevel] == 0:
            continue
        # Loop over cpu domains. In most cases, we'll only read the CPU corresponding to the domain.
        for ii in range(ncpu_selected):
            icpu = cpu_list[ii]
            nc = level_count[icpu, ilevel]
            if nc == 0:
                continue
            offset = offsets[icpu, ilevel]
            if offset == -1:
                continue
            f.seek(offset + skip_len(first_field_index, nc), SEEK_SET)

            # We have already skipped the first fields (if any)
            # so we "rewind" (this will cancel the first seek)
            jump_len = -first_field_index
            for i in range(twotondim):
                # Read the selected fields
                for j in range(nfields_selected):
                    jump_len += jumps[j]
                    if jump_len > 0:
                        f.seek(skip_len(jump_len, nc), SEEK_CUR)
                        jump_len = 0
                    f.read_vector_inplace('d', <void*> &buffer[0, i, j])

                jump_len += jumps[nfields_selected]

            # In principle, we may be left with some fields to skip
            # but since we're doing an absolute seek at the beginning of
            # the loop on CPUs, we can spare one seek here
            ## if jump_len > 0:
            ##     f.seek(skip_len(jump_len, nc), SEEK_CUR)

            # Alias buffer into dictionary
            tmp = {}
            for i, field in enumerate(fields):
                tmp[field] = buffer[:, :, i]

            if ncpu_selected > 1:
                oct_handler.fill_level_with_domain(
                    ilevel, level_inds, cell_inds, file_inds, domain_inds, tr, tmp, domain=icpu+1)
            else:
                oct_handler.fill_level(
                    ilevel, level_inds, cell_inds, file_inds, tr, tmp)



from yt._maintenance.deprecation import issue_deprecation_warning

from .unions import ParticleUnion  # noqa: F401

issue_deprecation_warning(
    "Importing ParticleUnion from yt.data_objects.particle_unions is deprecated. "
    "Please import this class from yt.data_objects.unions instead",
    stacklevel=3,
    since="4.2",
)



from numpy.testing import assert_allclose, assert_equal

from yt.utilities.chemical_formulas import ChemicalFormula, compute_mu
from yt.utilities.periodic_table import periodic_table

_molecules = (
    ("H2O_p1", (("H", 2), ("O", 1)), 1),
    ("H2O_m1", (("H", 2), ("O", 1)), -1),
    ("H2O", (("H", 2), ("O", 1)), 0),
    ("H2SO4", (("H", 2), ("S", 1), ("O", 4)), 0),
    # Now a harder one
    ("UuoMtUuq3", (("Uuo", 1), ("Mt", 1), ("Uuq", 3)), 0),
)


def test_formulas():
    for formula, components, charge in _molecules:
        f = ChemicalFormula(formula)
        w = sum(n * periodic_table[e].weight for e, n in components)
        assert_equal(f.charge, charge)
        assert_equal(f.weight, w)
        for (n, c1), (e, c2) in zip(components, f.elements, strict=True):
            assert_equal(n, e.symbol)
            assert_equal(c1, c2)


def test_default_mu():
    assert_allclose(compute_mu(None), 0.5924489101195808)
    assert_allclose(compute_mu("ionized"), 0.5924489101195808)
    assert_allclose(compute_mu("neutral"), 1.2285402715185552)



from . import tests
from .data_structures import (
    ExodusIIDataset,
    ExodusIIUnstructuredIndex,
    ExodusIIUnstructuredMesh,
)
from .fields import ExodusIIFieldInfo
from .io import IOHandlerExodusII
from .simulation_handling import ExodusIISimulation



import os

import yt.units
from yt.frontends.gadget.data_structures import GadgetHDF5Dataset
from yt.utilities.definitions import sec_conversion
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import OWLSFieldInfo


class OWLSDataset(GadgetHDF5Dataset):
    _load_requirements = ["h5py"]
    _particle_mass_name = "Mass"
    _field_info_class = OWLSFieldInfo
    _time_readin = "Time_GYR"

    def _parse_parameter_file(self):
        # read values from header
        hvals = self._get_hvals()
        self.parameters = hvals

        # set features common to OWLS and Eagle
        self._set_owls_eagle()

        # Set time from value in header
        self.current_time = (
            hvals[self._time_readin] * sec_conversion["Gyr"] * yt.units.s
        )

    def _set_code_unit_attributes(self):
        self._set_owls_eagle_units()

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        need_groups = ["Constants", "Header", "Parameters", "Units"]
        veto_groups = [
            "SUBFIND",
            "FOF",
            "PartType0/ChemistryAbundances",
            "PartType0/ChemicalAbundances",
            "RuntimePars",
            "HashTable",
        ]
        valid = True
        valid_fname = filename
        # If passed arg is a directory, look for the .0 file in that dir
        if os.path.isdir(filename):
            valid_files = []
            for f in os.listdir(filename):
                fname = os.path.join(filename, f)
                fext = os.path.splitext(fname)[-1]
                if (
                    (".0" in f)
                    and (fext not in {".ewah", ".kdtree"})
                    and os.path.isfile(fname)
                ):
                    valid_files.append(fname)
            if len(valid_files) == 0:
                valid = False
            elif len(valid_files) > 1:
                valid = False
            else:
                valid_fname = valid_files[0]
        try:
            fileh = h5py.File(valid_fname, mode="r")
            for ng in need_groups:
                if ng not in fileh["/"]:
                    valid = False
            for vg in veto_groups:
                if vg in fileh["/"]:
                    valid = False
            fileh.close()
        except Exception:
            valid = False
        return valid



from contextlib import contextmanager

from yt._maintenance.deprecation import issue_deprecation_warning
from yt.utilities.on_demand_imports import NotAModule, _h5py as h5py


def valid_hdf5_signature(fn: str, /) -> bool:
    signature = b"\x89HDF\r\n\x1a\n"
    try:
        with open(fn, "rb") as f:
            header = f.read(8)
            return header == signature
    except Exception:
        return False


def warn_h5py(fn):
    issue_deprecation_warning(
        "warn_h5py is not used within yt any more and is deprecated.", since="4.3"
    )
    needs_h5py = valid_hdf5_signature(fn)
    if needs_h5py and isinstance(h5py.File, NotAModule):
        raise RuntimeError(
            "This appears to be an HDF5 file, but h5py is not installed."
        )


class HDF5FileHandler:
    handle = None

    def __init__(self, filename):
        self.handle = h5py.File(filename, mode="r")

    def __getitem__(self, key):
        return self.handle[key]

    def __contains__(self, item):
        return item in self.handle

    def __len__(self):
        return len(self.handle)

    @property
    def attrs(self):
        return self.handle.attrs

    def keys(self):
        return list(self.handle.keys())

    def items(self):
        return list(self.handle.items())

    def close(self):
        if self.handle is not None:
            self.handle.close()


class FITSFileHandler(HDF5FileHandler):
    def __init__(self, filename):
        from yt.utilities.on_demand_imports import _astropy

        if isinstance(filename, _astropy.pyfits.hdu.image._ImageBaseHDU):
            self.handle = _astropy.pyfits.HDUList(filename)
        elif isinstance(filename, _astropy.pyfits.HDUList):
            self.handle = filename
        else:
            self.handle = _astropy.pyfits.open(
                filename, memmap=True, do_not_scale_image_data=True, ignore_blank=True
            )
        self._fits_files = []

    def __del__(self):
        for f in self._fits_files:
            f.close()
        del self._fits_files
        del self.handle
        self.handle = None

    def close(self):
        self.handle.close()


def valid_netcdf_signature(fn: str, /) -> bool:
    try:
        with open(fn, "rb") as f:
            header = f.read(4)
    except Exception:
        return False
    else:
        return header in (b"CDF\x01", b"CDF\x02") or (
            fn.endswith((".nc", ".nc4")) and valid_hdf5_signature(fn)
        )


def valid_netcdf_classic_signature(filename):
    issue_deprecation_warning(
        "valid_netcdf_classic_signature is not used within yt any more and is deprecated.",
        since="4.3",
    )
    signature_v1 = b"CDF\x01"
    signature_v2 = b"CDF\x02"
    try:
        with open(filename, "rb") as f:
            header = f.read(4)
            return header == signature_v1 or header == signature_v2
    except Exception:
        return False


def warn_netcdf(fn):
    # There are a few variants of the netCDF format.
    issue_deprecation_warning(
        "warn_netcdf is not used within yt any more and is deprecated.", since="4.3"
    )

    classic = valid_netcdf_classic_signature(fn)
    # NetCDF-4 Classic files are HDF5 files constrained to the Classic
    # data model used by netCDF-3.
    netcdf4_classic = valid_hdf5_signature(fn) and fn.endswith((".nc", ".nc4"))
    needs_netcdf = classic or netcdf4_classic
    from yt.utilities.on_demand_imports import _netCDF4 as netCDF4

    if needs_netcdf and isinstance(netCDF4.Dataset, NotAModule):
        raise RuntimeError(
            "This appears to be a netCDF file, but the "
            "python bindings for netCDF4 are not installed."
        )


class NetCDF4FileHandler:
    def __init__(self, filename):
        self.filename = filename

    @contextmanager
    def open_ds(self, **kwargs):
        from yt.utilities.on_demand_imports import _netCDF4 as netCDF4

        ds = netCDF4.Dataset(self.filename, mode="r", **kwargs)
        yield ds
        ds.close()



import os
import os.path
import shutil
import tempfile
from unittest import TestCase

import numpy as np

from yt.testing import assert_fname, fake_random_ds
from yt.visualization.volume_rendering.api import (
    ColorTransferFunction,
    ProjectionTransferFunction,
)
from yt.visualization.volume_rendering.old_camera import (
    FisheyeCamera,
    InteractiveCamera,
    PerspectiveCamera,
    ProjectionCamera,
    StereoPairCamera,
)


def setup_module():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class CameraTest(TestCase):
    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True

    def setUp(self):
        if self.use_tmpdir:
            self.curdir = os.getcwd()
            # Perform I/O in safe place instead of yt main dir
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)
        else:
            self.curdir, self.tmpdir = None, None

        self.ds = fake_random_ds(64)
        self.c = self.ds.domain_center
        self.L = np.array([0.5, 0.5, 0.5])
        self.W = 1.5 * self.ds.domain_width
        self.N = 64
        self.field = ("gas", "density")

    def tearDown(self):
        if self.use_tmpdir:
            os.chdir(self.curdir)
            shutil.rmtree(self.tmpdir)

    def setup_transfer_function(self, camera_type):
        if camera_type in ["perspective", "camera", "stereopair", "interactive"]:
            mi, ma = self.ds.all_data().quantities["Extrema"](self.field)
            tf = ColorTransferFunction((mi, ma), grey_opacity=True)
            tf.map_to_colormap(mi, ma, scale=10.0, colormap="RdBu_r")
            return tf
        elif camera_type in ["healpix"]:
            return ProjectionTransferFunction()
        else:
            pass

    def test_camera(self):
        tf = self.setup_transfer_function("camera")
        cam = self.ds.camera(
            self.c, self.L, self.W, self.N, transfer_function=tf, log_fields=[False]
        )
        cam.snapshot("camera.png")
        assert_fname("camera.png")

        im = cam.snapshot()
        im = cam.draw_domain(im)
        cam.draw_coordinate_vectors(im)
        cam.draw_line(im, [0, 0, 0], [1, 1, 0])

    def test_data_source_camera(self):
        ds = self.ds
        tf = self.setup_transfer_function("camera")
        data_source = ds.sphere(ds.domain_center, ds.domain_width[0] * 0.5)

        cam = ds.camera(
            self.c,
            self.L,
            self.W,
            self.N,
            log_fields=[False],
            transfer_function=tf,
            data_source=data_source,
        )
        cam.snapshot("data_source_camera.png")
        assert_fname("data_source_camera.png")

    def test_perspective_camera(self):
        ds = self.ds
        tf = self.setup_transfer_function("camera")

        cam = PerspectiveCamera(
            self.c,
            self.L,
            self.W,
            self.N,
            ds=ds,
            transfer_function=tf,
            log_fields=[False],
        )
        cam.snapshot("perspective.png")
        assert_fname("perspective.png")

    def test_interactive_camera(self):
        ds = self.ds
        tf = self.setup_transfer_function("camera")

        cam = InteractiveCamera(
            self.c,
            self.L,
            self.W,
            self.N,
            ds=ds,
            transfer_function=tf,
            log_fields=[False],
        )
        del cam
        # Can't take a snapshot here since IC uses matplotlib.'

    def test_projection_camera(self):
        ds = self.ds

        cam = ProjectionCamera(self.c, self.L, self.W, self.N, ds=ds, field=self.field)
        cam.snapshot("projection.png")
        assert_fname("projection.png")

    def test_stereo_camera(self):
        ds = self.ds
        tf = self.setup_transfer_function("camera")

        cam = ds.camera(
            self.c, self.L, self.W, self.N, transfer_function=tf, log_fields=[False]
        )
        stereo_cam = StereoPairCamera(cam)
        # Take image
        cam1, cam2 = stereo_cam.split()
        cam1.snapshot(fn="stereo1.png")
        cam2.snapshot(fn="stereo2.png")
        assert_fname("stereo1.png")
        assert_fname("stereo2.png")

    def test_camera_movement(self):
        ds = self.ds
        tf = self.setup_transfer_function("camera")

        cam = ds.camera(
            self.c,
            self.L,
            self.W,
            self.N,
            transfer_function=tf,
            log_fields=[False],
            north_vector=[0.0, 0.0, 1.0],
        )
        cam.zoom(0.5)
        for snap in cam.zoomin(2.0, 3):
            snap
        for snap in cam.move_to(
            np.array(self.c) + 0.1, 3, final_width=None, exponential=False
        ):
            snap
        for snap in cam.move_to(
            np.array(self.c) - 0.1, 3, final_width=2.0 * self.W, exponential=False
        ):
            snap
        for snap in cam.move_to(
            np.array(self.c), 3, final_width=1.0 * self.W, exponential=True
        ):
            snap
        cam.rotate(np.pi / 10)
        cam.pitch(np.pi / 10)
        cam.yaw(np.pi / 10)
        cam.roll(np.pi / 10)
        for snap in cam.rotation(np.pi, 3, rot_vector=None):
            snap
        for snap in cam.rotation(np.pi, 3, rot_vector=np.random.random(3)):
            snap
        cam.snapshot("final.png")
        assert_fname("final.png")

    def test_fisheye(self):
        ds = self.ds
        tf = self.setup_transfer_function("camera")
        cam = FisheyeCamera(
            ds.domain_center,
            ds.domain_width[0],
            360.0,
            256,
            transfer_function=tf,
            ds=ds,
        )
        cam.snapshot("fisheye.png")



from functools import partial

from numpy.testing import assert_raises

from yt.data_objects.static_output import Dataset
from yt.units import YTQuantity
from yt.units.unit_registry import UnitRegistry

mock_quan = partial(YTQuantity, registry=UnitRegistry())


def test_schema_validation():
    valid_schemas = [
        {"length_unit": 1.0},
        {"length_unit": [1.0]},
        {"length_unit": (1.0,)},
        {"length_unit": int(1.0)},
        {"length_unit": (1.0, "m")},
        {"length_unit": [1.0, "m"]},
        {"length_unit": YTQuantity(1.0, "m")},
    ]

    for schema in valid_schemas:
        uo = Dataset._sanitize_units_override(schema)
        for v in uo.values():
            q = mock_quan(v)  # check that no error (TypeError) is raised
            q.to("pc")  # check that q is a length


def test_invalid_schema_detection():
    invalid_key_schemas = [
        {"len_unit": 1.0},  # plain invalid key
        {"lenght_unit": 1.0},  # typo
    ]
    for invalid_schema in invalid_key_schemas:
        assert_raises(ValueError, Dataset._sanitize_units_override, invalid_schema)

    invalid_val_schemas = [
        {"length_unit": [1, 1, 1]},  # len(val) > 2
        {"length_unit": [1, 1, 1, 1, 1]},  # "data type not understood" in unyt
    ]

    for invalid_schema in invalid_val_schemas:
        assert_raises(TypeError, Dataset._sanitize_units_override, invalid_schema)

    # 0 shouldn't make sense
    invalid_number_schemas = [
        {"length_unit": 0},
        {"length_unit": [0]},
        {"length_unit": (0,)},
        {"length_unit": (0, "cm")},
    ]
    for invalid_schema in invalid_number_schemas:
        assert_raises(ValueError, Dataset._sanitize_units_override, invalid_schema)


def test_typing_error_detection():
    invalid_schema = {"length_unit": "1m"}

    # this is the error that is raised by unyt on bad input
    assert_raises(RuntimeError, mock_quan, invalid_schema["length_unit"])

    # check that the sanitizer function is able to catch the
    # type issue before passing down to unyt
    assert_raises(TypeError, Dataset._sanitize_units_override, invalid_schema)


def test_dimensionality_error_detection():
    invalid_schema = {"length_unit": YTQuantity(1.0, "s")}
    assert_raises(ValueError, Dataset._sanitize_units_override, invalid_schema)



import pytest
from numpy.testing import assert_equal

import yt
from yt.testing import fake_random_ds
from yt.visualization.line_plot import _validate_point


def setup_module():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class TestLinePlotSimple:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_random_ds(4)
        fields = [field for field in cls.ds.field_list if field[0] == "stream"]
        field_labels = {f: f[1] for f in fields}
        plot = yt.LinePlot(
            cls.ds, fields, (0, 0, 0), (1, 1, 0), 1000, field_labels=field_labels
        )
        plot.annotate_legend(fields[0])
        plot.annotate_legend(fields[1])
        plot.set_x_unit("cm")
        plot.set_unit(fields[0], "kg/cm**3")
        plot.annotate_title(fields[0], "Density Plot")
        plot.render()
        cls.plot = plot

    @pytest.mark.mpl_image_compare
    def test_lineplot_simple_density(self):
        return self.plot.plots["stream", "density"].figure

    @pytest.mark.mpl_image_compare
    def test_lineplot_simple_velocity_x(self):
        return self.plot.plots["stream", "velocity_x"].figure


class TestLinePlotMulti:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_random_ds(4)
        fields = [field for field in cls.ds.field_list if field[0] == "stream"]
        field_labels = {f: f[1] for f in fields}
        lines = []
        lines.append(
            yt.LineBuffer(cls.ds, [0.25, 0, 0], [0.25, 1, 0], 100, label="x = 0.5")
        )
        lines.append(
            yt.LineBuffer(cls.ds, [0.5, 0, 0], [0.5, 1, 0], 100, label="x = 0.5")
        )
        plot = yt.LinePlot.from_lines(cls.ds, fields, lines, field_labels=field_labels)
        plot.annotate_legend(fields[0])
        plot.annotate_legend(fields[1])
        plot.set_x_unit("cm")
        plot.set_unit(fields[0], "kg/cm**3")
        plot.annotate_title(fields[0], "Density Plot")
        plot.render()
        cls.plot = plot

    @pytest.mark.mpl_image_compare
    def test_lineplot_multi_density(self):
        return self.plot.plots["stream", "density"].figure

    @pytest.mark.mpl_image_compare
    def test_lineplot_multi_velocity_x(self):
        return self.plot.plots["stream", "velocity_x"].figure


def test_line_buffer():
    ds = fake_random_ds(32)
    lb = yt.LineBuffer(ds, (0, 0, 0), (1, 1, 1), 512, label="diag")
    lb["gas", "density"]
    lb["gas", "velocity_x"]
    assert_equal(lb["gas", "density"].size, 512)
    lb["gas", "density"] = 0
    assert_equal(lb["gas", "density"], 0)
    assert_equal(set(lb.keys()), {("gas", "density"), ("gas", "velocity_x")})
    del lb["gas", "velocity_x"]
    assert_equal(set(lb.keys()), {("gas", "density")})


def test_validate_point():
    ds = fake_random_ds(3)
    with pytest.raises(RuntimeError, match="Input point must be array-like"):
        _validate_point(0, ds, start=True)

    with pytest.raises(RuntimeError, match="Input point must be a 1D array"):
        _validate_point(ds.arr([[0], [1]], "code_length"), ds, start=True)

    with pytest.raises(
        RuntimeError, match="Input point must have an element for each dimension"
    ):
        _validate_point(ds.arr([0, 1], "code_length"), ds, start=True)

    ds = fake_random_ds([32, 32, 1])
    _validate_point(ds.arr([0, 1], "code_length"), ds, start=True)
    _validate_point(ds.arr([0, 1], "code_length"), ds)



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer


class FITSFieldInfo(FieldInfoContainer):
    known_other_fields = ()

    def __init__(self, ds, field_list, slice_info=None):
        super().__init__(ds, field_list, slice_info=slice_info)
        for field in ds.field_list:
            if field[0] == "fits":
                self[field].take_log = False


class YTFITSFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("density", ("code_mass/code_length**3", ["density"], None)),
        (
            "dark_matter_density",
            ("code_mass/code_length**3", ["dark_matter_density"], None),
        ),
        ("number_density", ("1/code_length**3", ["number_density"], None)),
        ("pressure", ("dyne/code_length**2", ["pressure"], None)),
        ("thermal_energy", ("erg / g", ["specific_thermal_energy"], None)),
        ("temperature", ("K", ["temperature"], None)),
        ("velocity_x", ("code_length/code_time", ["velocity_x"], None)),
        ("velocity_y", ("code_length/code_time", ["velocity_y"], None)),
        ("velocity_z", ("code_length/code_time", ["velocity_z"], None)),
        ("magnetic_field_x", ("gauss", [], None)),
        ("magnetic_field_y", ("gauss", [], None)),
        ("magnetic_field_z", ("gauss", [], None)),
        ("metallicity", ("Zsun", ["metallicity"], None)),
        # We need to have a bunch of species fields here, too
        ("metal_density", ("code_mass/code_length**3", ["metal_density"], None)),
        ("hi_density", ("code_mass/code_length**3", ["hi_density"], None)),
        ("hii_density", ("code_mass/code_length**3", ["hii_density"], None)),
        ("h2i_density", ("code_mass/code_length**3", ["h2i_density"], None)),
        ("h2ii_density", ("code_mass/code_length**3", ["h2ii_density"], None)),
        ("h2m_density", ("code_mass/code_length**3", ["h2m_density"], None)),
        ("hei_density", ("code_mass/code_length**3", ["hei_density"], None)),
        ("heii_density", ("code_mass/code_length**3", ["heii_density"], None)),
        ("heiii_density", ("code_mass/code_length**3", ["heiii_density"], None)),
        ("hdi_density", ("code_mass/code_length**3", ["hdi_density"], None)),
        ("di_density", ("code_mass/code_length**3", ["di_density"], None)),
        ("dii_density", ("code_mass/code_length**3", ["dii_density"], None)),
    )

    def __init__(self, ds, field_list, slice_info=None):
        super().__init__(ds, field_list, slice_info=slice_info)


class WCSFITSFieldInfo(FITSFieldInfo):
    def setup_fluid_fields(self):
        wcs_2d = getattr(self.ds, "wcs_2d", self.ds.wcs)

        def _pixel(field, data):
            return data.ds.arr(data["index", "ones"], "pixel")

        self.add_field(
            ("fits", "pixel"), sampling_type="cell", function=_pixel, units="pixel"
        )

        def _get_2d_wcs(data, axis):
            w_coords = wcs_2d.wcs_pix2world(data["index", "x"], data["index", "y"], 1)
            return w_coords[axis]

        def world_f(axis, unit):
            def _world_f(field, data):
                return data.ds.arr(_get_2d_wcs(data, axis), unit)

            return _world_f

        for i, axis, name in [
            (0, self.ds.lon_axis, self.ds.lon_name),
            (1, self.ds.lat_axis, self.ds.lat_name),
        ]:
            unit = str(wcs_2d.wcs.cunit[i])
            if unit.lower() == "deg":
                unit = "degree"
            if unit.lower() == "rad":
                unit = "radian"
            self.add_field(
                ("fits", name),
                sampling_type="cell",
                function=world_f(axis, unit),
                units=unit,
            )

        if self.ds.dimensionality == 3:

            def _spec(field, data):
                axis = "xyz"[data.ds.spec_axis]
                sp = (
                    data["fits", axis].ndarray_view() - self.ds._p0
                ) * self.ds._dz + self.ds._z0
                return data.ds.arr(sp, data.ds.spec_unit)

            self.add_field(
                ("fits", "spectral"),
                sampling_type="cell",
                function=_spec,
                units=self.ds.spec_unit,
                display_name=self.ds.spec_name,
            )



import numpy as np

_index_map = np.array(
    [
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 1, 0, 1, 0, 1],
        [0, 0, 1, 1, 0, 0, 1, 1],
        [0, 1, 2, 3, 0, 1, 2, 3],
        [0, 0, 0, 0, 1, 1, 1, 1],
        [0, 1, 0, 1, 2, 3, 2, 3],
        [0, 0, 1, 1, 2, 2, 3, 3],
        [0, 1, 2, 3, 4, 5, 6, 7],
    ]
)


def _get_linear_index(nodal_flag):
    if len(nodal_flag) == 2:
        return 1 * nodal_flag[1] + 2 * nodal_flag[0]
    else:
        return 1 * nodal_flag[2] + 2 * nodal_flag[1] + 4 * nodal_flag[0]


def _get_indices(nodal_flag):
    li = _get_linear_index(nodal_flag)
    return _index_map[li]


def get_nodal_data(data_source, field):
    finfo = data_source.ds._get_field_info(field)
    nodal_flag = finfo.nodal_flag
    field_data = data_source[field]
    inds = _get_indices(nodal_flag)
    return field_data[:, inds]


def get_nodal_slices(shape, nodal_flag, dim):
    slices = []
    dir_slices = [[] for _ in range(dim)]

    for i in range(dim):
        if nodal_flag[i]:
            dir_slices[i] = [slice(0, shape[i] - 1), slice(1, shape[i])]
        else:
            dir_slices[i] = [slice(0, shape[i])]

    for sl_i in dir_slices[0]:
        for sl_j in dir_slices[1]:
            if dim > 2:
                for sl_k in dir_slices[2]:
                    slices.append([sl_i, sl_j, sl_k])
            else:
                slices.append([sl_i, sl_j])

    return tuple(slices)



from ...amrex.data_structures import (
    AMReXDataset,
    AMReXHierarchy,
    BoxlibDataset,
    BoxlibGrid,
    BoxlibHierarchy,
    CastroDataset,
    MaestroDataset,
    NyxDataset,
    NyxHierarchy,
    OrionDataset,
    OrionHierarchy,
    WarpXDataset,
    WarpXHierarchy,
)
from .._deprecation import boxlib_deprecation

boxlib_deprecation()



import os.path

from numpy.testing import assert_equal

from yt.testing import requires_module
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    data_dir_load,
    requires_ds,
)

# from yt.frontends.owls_subfind.api import OWLSSubfindDataset

_fields = (
    ("all", "particle_position_x"),
    ("all", "particle_position_y"),
    ("all", "particle_position_z"),
    ("all", "particle_mass"),
)

# a dataset with empty files
g1 = "owls_fof_halos/groups_001/group_001.0.hdf5"
g8 = "owls_fof_halos/groups_008/group_008.0.hdf5"


@requires_module("h5py")
@requires_ds(g8)
def test_fields_g8():
    ds = data_dir_load(g8)
    assert_equal(str(ds), os.path.basename(g8))
    for field in _fields:
        yield FieldValuesTest(g8, field, particle_type=True)


@requires_module("h5py")
@requires_ds(g1)
def test_fields_g1():
    ds = data_dir_load(g1)
    assert_equal(str(ds), os.path.basename(g1))
    for field in _fields:
        yield FieldValuesTest(g1, field, particle_type=True)



import abc
import base64
import os
import warnings
from collections import defaultdict
from functools import wraps
from typing import Any, Final, Literal

import matplotlib
from matplotlib.colors import LogNorm, Normalize, SymLogNorm
from unyt.dimensions import length

from yt._maintenance.deprecation import issue_deprecation_warning
from yt._maintenance.ipython_compat import IS_IPYTHON
from yt._typing import FieldKey, Quantity
from yt.config import ytcfg
from yt.data_objects.time_series import DatasetSeries
from yt.funcs import ensure_dir, is_sequence, iter_fields
from yt.units.unit_object import Unit  # type: ignore
from yt.utilities.definitions import formatted_length_unit_names
from yt.utilities.exceptions import YTConfigurationError, YTNotInsideNotebook
from yt.visualization._commons import get_default_from_config
from yt.visualization._handlers import ColorbarHandler, NormHandler
from yt.visualization.base_plot_types import PlotMPL

from ._commons import (
    _get_units_label,
    get_default_font_properties,
    invalidate_data,
    invalidate_figure,
    invalidate_plot,
    validate_image_name,
    validate_plot,
)

latex_prefixes = {
    "u": r"\mu",
}


def apply_callback(f):
    issue_deprecation_warning(
        "The apply_callback decorator is not used in yt any more and "
        "will be removed in a future version. "
        "Please do not use it.",
        stacklevel=3,
        since="4.1",
    )

    @wraps(f)
    def newfunc(*args, **kwargs):
        args[0]._callbacks.append((f.__name__, (args, kwargs)))
        return args[0]

    return newfunc


def accepts_all_fields(func):
    """
    Decorate a function whose second argument is <field> and deal with the special case
    field == 'all', looping over all fields already present in the PlotContainer object.

    """

    # This is to be applied to PlotContainer class methods with the following signature:
    #
    # f(self, field, *args, **kwargs) -> self
    @wraps(func)
    def newfunc(self, field, *args, **kwargs):
        if field == "all":
            field = self.plots.keys()
        for f in self.data_source._determine_fields(field):
            func(self, f, *args, **kwargs)
        return self

    return newfunc


# define a singleton sentinel to be used as default value distinct from None
class Unset:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance


UNSET: Final = Unset()


class PlotDictionary(defaultdict):
    def __getitem__(self, item):
        return defaultdict.__getitem__(
            self, self.data_source._determine_fields(item)[0]
        )

    def __setitem__(self, item, value):
        return defaultdict.__setitem__(
            self, self.data_source._determine_fields(item)[0], value
        )

    def __contains__(self, item):
        return defaultdict.__contains__(
            self, self.data_source._determine_fields(item)[0]
        )

    def __init__(self, data_source, default_factory=None):
        self.data_source = data_source
        return defaultdict.__init__(self, default_factory)


class PlotContainer(abc.ABC):
    """A container for generic plots"""

    _plot_dict_type: type[PlotDictionary] = PlotDictionary
    _plot_type: str | None = None
    _plot_valid = False

    _default_figure_size = tuple(matplotlib.rcParams["figure.figsize"])
    _default_font_size = 14.0

    def __init__(self, data_source, figure_size=None, fontsize: float | None = None):
        from matplotlib.font_manager import FontProperties

        self.data_source = data_source
        self.ds = data_source.ds
        self.ts = self._initialize_dataset(self.ds)
        self.plots = self.__class__._plot_dict_type(data_source)

        self._set_figure_size(figure_size)

        if fontsize is None:
            fontsize = self.__class__._default_font_size
        font_dict = get_default_font_properties() | {"size": fontsize}

        self._font_properties = FontProperties(**font_dict)
        self._font_color = None
        self._xlabel = None
        self._ylabel = None
        self._minorticks: dict[FieldKey, bool] = {}

    @accepts_all_fields
    @invalidate_plot
    def set_log(
        self,
        field,
        log: bool | None = None,
        *,
        linthresh: float | Quantity | Literal["auto"] | None = None,
        symlog_auto: bool | None = None,  # deprecated
    ):
        """set a field to log, linear, or symlog.

        Symlog scaling is a combination of linear and log, where from 0 to a
        threshold value, it operates as linear, and then beyond that it operates as
        log.  Symlog can also work with negative values in log space as well as
        negative and positive values simultaneously and symmetrically.  If symlog
        scaling is desired, please set log=True and either set symlog_auto=True or
        select a value for linthresh.

        Parameters
        ----------
        field : string
            the field to set a transform
            if field == 'all', applies to all plots.
        log : boolean, optional
            set log to True for log scaling, False for linear scaling.
        linthresh : float, (float, str), unyt_quantity, or 'auto', optional
            when using symlog scaling, linthresh is the value at which scaling
            transitions from linear to logarithmic.  linthresh must be positive.
            Note: setting linthresh will automatically enable symlog scale

        Note that *log* and *linthresh* are mutually exclusive arguments
        """
        if log is None and linthresh is None and symlog_auto is None:
            raise TypeError("set_log requires log or linthresh be set")

        if symlog_auto is not None:
            issue_deprecation_warning(
                "the symlog_auto argument is deprecated. Use linthresh='auto' instead",
                since="4.1",
                stacklevel=5,
            )
            if symlog_auto is True:
                linthresh = "auto"
            elif symlog_auto is False:
                pass
            else:
                raise TypeError(
                    "Received invalid value for parameter symlog_auto. "
                    f"Expected a boolean, got {symlog_auto!r}"
                )

        if log is not None and linthresh is not None:
            # we do not raise an error here for backward compatibility
            warnings.warn(
                f"log={log} has no effect because linthresh specified. Using symlog.",
                stacklevel=4,
            )

        pnh = self.plots[field].norm_handler

        if linthresh is not None:
            if isinstance(linthresh, str):
                if linthresh == "auto":
                    pnh.norm_type = SymLogNorm
                else:
                    raise ValueError(
                        "Expected a number, a unyt_quantity, a (float, 'unit') tuple, or 'auto'. "
                        f"Got linthresh={linthresh!r}"
                    )
            else:
                # pnh takes care of switching to symlog when linthresh is set
                pnh.linthresh = linthresh
        elif log is True:
            pnh.norm_type = LogNorm
        elif log is False:
            pnh.norm_type = Normalize
        else:
            raise TypeError(
                f"Could not parse arguments log={log!r}, linthresh={linthresh!r}"
            )

        return self

    def get_log(self, field):
        """get the transform type of a field.

        Parameters
        ----------
        field : string
            the field to get a transform
            if field == 'all', applies to all plots.

        """
        # devnote : accepts_all_fields decorator is not applicable here because
        # the return variable isn't self
        issue_deprecation_warning(
            "The get_log method is not reliable and is deprecated. "
            "Please do not rely on it.",
            stacklevel=3,
            since="4.1",
        )
        log = {}
        if field == "all":
            fields = list(self.plots.keys())
        else:
            fields = field
        for field in self.data_source._determine_fields(fields):
            pnh = self.plots[field].norm_handler
            if pnh.norm is not None:
                log[field] = type(pnh.norm) is LogNorm
            elif pnh.norm_type is not None:
                log[field] = pnh.norm_type is LogNorm
            else:
                # the NormHandler object has no constraints yet
                # so we'll assume defaults
                log[field] = True
        return log

    @invalidate_plot
    def set_transform(self, field, name: str):
        field = self.data_source._determine_fields(field)[0]
        pnh = self.plots[field].norm_handler
        pnh.norm_type = {
            "linear": Normalize,
            "log10": LogNorm,
            "symlog": SymLogNorm,
        }[name]
        return self

    @accepts_all_fields
    @invalidate_plot
    def set_norm(self, field, norm: Normalize):
        r"""
        Set a custom ``matplotlib.colors.Normalize`` to plot *field*.

        Any constraints previously set with `set_log`, `set_zlim` will be
        dropped.

        Note that any float value attached to *norm* (e.g. vmin, vmax,
        vcenter ...) will be read in the current displayed units, which can be
        controlled with the `set_unit` method.

        Parameters
        ----------
        field : str or tuple[str, str]
            if field == 'all', applies to all plots.
        norm : matplotlib.colors.Normalize
            see https://matplotlib.org/stable/tutorials/colors/colormapnorms.html
        """
        pnh = self.plots[field].norm_handler
        pnh.norm = norm
        return self

    @accepts_all_fields
    @invalidate_plot
    def set_minorticks(self, field, state):
        """Turn minor ticks on or off in the current plot.

        Displaying minor ticks reduces performance; turn them off
        using set_minorticks('all', False) if drawing speed is a problem.

        Parameters
        ----------
        field : string
            the field to remove minorticks
            if field == 'all', applies to all plots.
        state : bool
            the state indicating 'on' (True) or 'off' (False)

        """
        self._minorticks[field] = state
        return self

    @abc.abstractmethod
    def _setup_plots(self):
        # Left blank to be overridden in subclasses
        pass

    def render(self) -> None:
        r"""Render plots.
        This operation is expensive and usually doesn't need to be requested explicitly.
        In most cases, yt handles rendering automatically and delays it as much as possible
        to avoid redundant calls on each plot modification (e.g. via `annotate_*` methods).

        However, valid use cases of this method include:
        - fine control of render (and clear) operations when yt plots are combined with plot
          customizations other than plot callbacks (`annotate_*`)
        - testing
        """
        # this public API method should never be no-op, so we invalidate
        # the plot to force a fresh render in _setup_plots()
        self._plot_valid = False
        self._setup_plots()

    def _initialize_dataset(self, ts):
        if not isinstance(ts, DatasetSeries):
            if not is_sequence(ts):
                ts = [ts]
            ts = DatasetSeries(ts)
        return ts

    @invalidate_data
    def _switch_ds(self, new_ds, data_source=None):
        old_object = self.data_source
        name = old_object._type_name
        kwargs = {n: getattr(old_object, n) for n in old_object._con_args}
        kwargs["center"] = getattr(old_object, "center", None)
        if data_source is not None:
            if name != "proj":
                raise RuntimeError(
                    "The data_source keyword argument "
                    "is only defined for projections."
                )
            kwargs["data_source"] = data_source

        self.ds = new_ds

        # A _hack_ for ParticleProjectionPlots
        if name == "Particle":
            from yt.visualization.particle_plots import (
                ParticleAxisAlignedDummyDataSource,
            )

            new_object = ParticleAxisAlignedDummyDataSource(ds=self.ds, **kwargs)
        else:
            new_object = getattr(new_ds, name)(**kwargs)

        self.data_source = new_object

        for d in "xyz":
            lim_name = d + "lim"
            if hasattr(self, lim_name):
                lim = getattr(self, lim_name)
                lim = tuple(new_ds.quan(l.value, str(l.units)) for l in lim)
                setattr(self, lim_name, lim)
        self.plots.data_source = new_object
        self._colorbar_label.data_source = new_object
        self._setup_plots()

    @validate_plot
    def __getitem__(self, item):
        return self.plots[item]

    def _set_font_properties(self):
        for f in self.plots:
            self.plots[f]._set_font_properties(self._font_properties, self._font_color)

    @invalidate_plot
    @invalidate_figure
    def set_font(self, font_dict=None):
        """

        Set the font and font properties.

        Parameters
        ----------

        font_dict : dict
            A dict of keyword parameters to be passed to
            :class:`matplotlib.font_manager.FontProperties`.

            Possible keys include:

            * family - The font family. Can be serif, sans-serif, cursive,
              'fantasy' or 'monospace'.
            * style - The font style. Either normal, italic or oblique.
            * color - A valid color string like 'r', 'g', 'red', 'cobalt',
              and 'orange'.
            * variant - Either normal or small-caps.
            * size - Either a relative value of xx-small, x-small, small,
              medium, large, x-large, xx-large or an absolute font size, e.g. 12
            * stretch - A numeric value in the range 0-1000 or one of
              ultra-condensed, extra-condensed, condensed, semi-condensed,
              normal, semi-expanded, expanded, extra-expanded or ultra-expanded
            * weight - A numeric value in the range 0-1000 or one of ultralight,
              light, normal, regular, book, medium, roman, semibold, demibold,
              demi, bold, heavy, extra bold, or black

            See the matplotlib font manager API documentation for more details.
            https://matplotlib.org/stable/api/font_manager_api.html

        Notes
        -----

        Mathtext axis labels will only obey the `size` and `color` keyword.

        Examples
        --------

        This sets the font to be 24-pt, blue, sans-serif, italic, and
        bold-face.

        >>> slc = SlicePlot(ds, "x", "Density")
        >>> slc.set_font(
        ...     {
        ...         "family": "sans-serif",
        ...         "style": "italic",
        ...         "weight": "bold",
        ...         "size": 24,
        ...         "color": "blue",
        ...     }
        ... )

        """
        from matplotlib.font_manager import FontProperties

        if font_dict is None:
            font_dict = {}
        if "color" in font_dict:
            self._font_color = font_dict.pop("color")
        # Set default values if the user does not explicitly set them.
        # this prevents reverting to the matplotlib defaults.
        _default_size = {"size": self.__class__._default_font_size}
        font_dict = get_default_font_properties() | _default_size | font_dict
        self._font_properties = FontProperties(**font_dict)
        return self

    def set_font_size(self, size):
        """Set the size of the font used in the plot

        This sets the font size by calling the set_font function.  See set_font
        for more font customization options.

        Parameters
        ----------
        size : float
        The absolute size of the font in points (1 pt = 1/72 inch).

        """
        return self.set_font({"size": size})

    def _set_figure_size(self, size):
        if size is None:
            self.figure_size = self.__class__._default_figure_size
        elif is_sequence(size):
            if len(size) != 2:
                raise TypeError(f"Expected a single float or a pair, got {size}")
            self.figure_size = float(size[0]), float(size[1])
        else:
            self.figure_size = float(size)

    @invalidate_plot
    @invalidate_figure
    def set_figure_size(self, size):
        """Sets a new figure size for the plot

        parameters
        ----------
        size : float, a sequence of two floats, or None
            The size of the figure (in units of inches),  including the margins
            but not the colorbar. If a single float is passed, it's interpreted
            as the size along the long axis.
            Pass None to reset
        """
        self._set_figure_size(size)
        return self

    @validate_plot
    def save(
        self,
        name: str | list[str] | tuple[str, ...] | None = None,
        suffix: str | None = None,
        mpl_kwargs: dict[str, Any] | None = None,
    ):
        """saves the plot to disk.

        Parameters
        ----------
        name : string or tuple, optional
           The base of the filename. If name is a directory or if name is not
           set, the filename of the dataset is used. For a tuple, the
           resulting path will be given by joining the elements of the
           tuple
        suffix : string, optional
           Specify the image type by its suffix. If not specified, the output
           type will be inferred from the filename. Defaults to '.png'.
        mpl_kwargs : dict, optional
           A dict of keyword arguments to be passed to matplotlib.

        >>> slc.save(mpl_kwargs={"bbox_inches": "tight"})

        """
        names = []
        if mpl_kwargs is None:
            mpl_kwargs = {}
        elif "format" in mpl_kwargs:
            new_suffix = mpl_kwargs.pop("format")
            if new_suffix != suffix:
                warnings.warn(
                    f"Overriding suffix {suffix!r} with mpl_kwargs['format'] = {new_suffix!r}. "
                    "Use the `suffix` argument directly to suppress this warning.",
                    stacklevel=2,
                )
            suffix = new_suffix

        if name is None:
            name = str(self.ds)
        elif isinstance(name, (list, tuple)):
            if not all(isinstance(_, str) for _ in name):
                raise TypeError(
                    f"Expected a single str or an iterable of str, got {name!r}"
                )
            name = os.path.join(*name)

        name = os.path.expanduser(name)

        parent_dir, _, prefix1 = name.replace(os.sep, "/").rpartition("/")
        parent_dir = parent_dir.replace("/", os.sep)

        if parent_dir and not os.path.isdir(parent_dir):
            ensure_dir(parent_dir)

        if name.endswith(("/", os.path.sep)):
            name = os.path.join(name, str(self.ds))

        new_name = validate_image_name(name, suffix)
        if new_name == name:
            for v in self.plots.values():
                out_name = v.save(name, mpl_kwargs)
                names.append(out_name)
            return names

        name = new_name
        prefix, suffix = os.path.splitext(name)

        if hasattr(self.data_source, "axis"):
            axis = self.ds.coordinates.axis_name.get(self.data_source.axis, "")
        else:
            axis = None
        weight = None
        stddev = None
        plot_type = self._plot_type
        if plot_type in ["Projection", "OffAxisProjection"]:
            weight = self.data_source.weight_field
            if weight is not None:
                weight = weight[1].replace(" ", "_")
            if getattr(self.data_source, "moment", 1) == 2:
                stddev = "standard_deviation"
        if "Cutting" in self.data_source.__class__.__name__:
            plot_type = "OffAxisSlice"

        for k, v in self.plots.items():
            if isinstance(k, tuple):
                k = k[1]

            if plot_type is None:
                # implemented this check to make mypy happy, because we can't use str.join
                # with PlotContainer._plot_type = None
                raise TypeError(f"{self.__class__} is missing a _plot_type value (str)")

            name_elements = [prefix, plot_type]
            if axis:
                name_elements.append(axis)
            name_elements.append(k.replace(" ", "_"))
            if weight:
                name_elements.append(weight)
            if stddev:
                name_elements.append(stddev)
            name = "_".join(name_elements) + suffix
            names.append(v.save(name, mpl_kwargs))
        return names

    @invalidate_data
    def refresh(self):
        # invalidate_data will take care of everything
        return self

    @validate_plot
    def show(self):
        r"""This will send any existing plots to the IPython notebook.

        If yt is being run from within an IPython session, and it is able to
        determine this, this function will send any existing plots to the
        notebook for display.

        If yt can't determine if it's inside an IPython session, it will raise
        YTNotInsideNotebook.

        Examples
        --------

        >>> from yt import SlicePlot
        >>> slc = SlicePlot(
        ...     ds, "x", [("gas", "density"), ("gas", "velocity_magnitude")]
        ... )
        >>> slc.show()

        """
        interactivity = self.plots[list(self.plots.keys())[0]].interactivity
        if interactivity:
            for v in sorted(self.plots.values()):
                v.show()
        else:
            if IS_IPYTHON:
                from IPython.display import display

                display(self)
            else:
                raise YTNotInsideNotebook

    @validate_plot
    def display(self, name=None, mpl_kwargs=None):
        """Will attempt to show the plot in in an IPython notebook.
        Failing that, the plot will be saved to disk."""
        try:
            return self.show()
        except YTNotInsideNotebook:
            return self.save(name=name, mpl_kwargs=mpl_kwargs)

    @validate_plot
    def _repr_html_(self):
        """Return an html representation of the plot object. Will display as a
        png for each WindowPlotMPL instance in self.plots"""
        ret = ""
        for field in self.plots:
            img = base64.b64encode(self.plots[field]._repr_png_()).decode()
            ret += (
                r'<img style="max-width:100%;max-height:100%;" '
                rf'src="data:image/png;base64,{img}"><br>'
            )
        return ret

    @invalidate_plot
    def set_xlabel(self, label):
        r"""
        Allow the user to modify the X-axis title
        Defaults to the global value. Fontsize defaults
        to 18.

        Parameters
        ----------
        label : str
            The new string for the x-axis.

        >>> plot.set_xlabel("H2I Number Density (cm$^{-3}$)")

        """
        self._xlabel = label
        return self

    @invalidate_plot
    def set_ylabel(self, label):
        r"""
        Allow the user to modify the Y-axis title
        Defaults to the global value.

        Parameters
        ----------
        label : str
            The new string for the y-axis.

        >>> plot.set_ylabel("Temperature (K)")

        """
        self._ylabel = label
        return self

    def _get_axes_unit_labels(self, unit_x, unit_y):
        axes_unit_labels = ["", ""]
        comoving = False
        hinv = False
        for i, un in enumerate((unit_x, unit_y)):
            unn = None
            if hasattr(self.data_source, "axis"):
                if hasattr(self.ds.coordinates, "image_units"):
                    # This *forces* an override
                    unn = self.ds.coordinates.image_units[self.data_source.axis][i]
                elif hasattr(self.ds.coordinates, "default_unit_label"):
                    axax = getattr(self.ds.coordinates, f"{'xy'[i]}_axis")[
                        self.data_source.axis
                    ]
                    unn = self.ds.coordinates.default_unit_label.get(axax, None)
            if unn in (1, "1", "dimensionless"):
                axes_unit_labels[i] = ""
                continue
            if unn is not None:
                axes_unit_labels[i] = _get_units_label(unn).strip("$")
                continue
            # Use sympy to factor h out of the unit.  In this context 'un'
            # is a string, so we call the Unit constructor.
            expr = Unit(un, registry=self.ds.unit_registry).expr
            h_expr = Unit("h", registry=self.ds.unit_registry).expr
            # See http://docs.sympy.org/latest/modules/core.html#sympy.core.expr.Expr
            h_power = expr.as_coeff_exponent(h_expr)[1]
            # un is now the original unit, but with h factored out.
            un = str(expr * h_expr ** (-1 * h_power))
            un_unit = Unit(un, registry=self.ds.unit_registry)
            cm = Unit("cm").expr
            if str(un).endswith("cm") and cm not in un_unit.expr.atoms():
                comoving = True
                un = un[:-2]
            # no length units besides code_length end in h so this is safe
            if h_power == -1:
                hinv = True
            elif h_power != 0:
                # It doesn't make sense to scale a position by anything
                # other than h**-1
                raise RuntimeError
            if un not in ["1", "u", "unitary"]:
                if un in formatted_length_unit_names:
                    un = formatted_length_unit_names[un]
                else:
                    un = Unit(un, registry=self.ds.unit_registry)
                    un = un.latex_representation()
                    if hinv:
                        un = un + r"\,h^{-1}"
                    if comoving:
                        un = un + r"\,(1+z)^{-1}"
                    pp = un[0]
                    if pp in latex_prefixes:
                        symbol_wo_prefix = un[1:]
                        if symbol_wo_prefix in self.ds.unit_registry.prefixable_units:
                            un = un.replace(pp, "{" + latex_prefixes[pp] + "}", 1)
                axes_unit_labels[i] = _get_units_label(un).strip("$")
        return axes_unit_labels

    def hide_colorbar(self, field=None):
        """
        Hides the colorbar for a plot and updates the size of the
        plot accordingly.  Defaults to operating on all fields for a
        PlotContainer object.

        Parameters
        ----------

        field : string, field tuple, or list of strings or field tuples (optional)
            The name of the field(s) that we want to hide the colorbar.
            If None or 'all' is provided, will default to using all fields available
            for this object.

        Examples
        --------

        This will save an image with no colorbar.

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> s = SlicePlot(ds, 2, "density", "c", (20, "kpc"))
        >>> s.hide_colorbar()
        >>> s.save()

        This will save an image with no axis or colorbar.

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> s = SlicePlot(ds, 2, "density", "c", (20, "kpc"))
        >>> s.hide_axes()
        >>> s.hide_colorbar()
        >>> s.save()
        """
        if field is None or field == "all":
            field = self.plots.keys()
        for f in self.data_source._determine_fields(field):
            self.plots[f].hide_colorbar()
        return self

    def show_colorbar(self, field=None):
        """
        Shows the colorbar for a plot and updates the size of the
        plot accordingly.  Defaults to operating on all fields for a
        PlotContainer object.  See hide_colorbar().

        Parameters
        ----------

        field : string, field tuple, or list of strings or field tuples (optional)
        The name of the field(s) that we want to show the colorbar.
        """
        if field is None:
            field = self.fields
        for f in iter_fields(field):
            self.plots[f].show_colorbar()
        return self

    def hide_axes(self, field=None, draw_frame=None):
        """
        Hides the axes for a plot and updates the size of the
        plot accordingly.  Defaults to operating on all fields for a
        PlotContainer object.

        Parameters
        ----------

        field : string, field tuple, or list of strings or field tuples (optional)
            The name of the field(s) that we want to hide the axes.

        draw_frame : boolean
            If True, the axes frame will still be drawn. Defaults to False.
            See note below for more details.

        Examples
        --------

        This will save an image with no axes.

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> s = SlicePlot(ds, 2, "density", "c", (20, "kpc"))
        >>> s.hide_axes()
        >>> s.save()

        This will save an image with no axis or colorbar.

        >>> import yt
        >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
        >>> s = SlicePlot(ds, 2, "density", "c", (20, "kpc"))
        >>> s.hide_axes()
        >>> s.hide_colorbar()
        >>> s.save()

        Note
        ----
        By default, when removing the axes, the patch on which the axes are
        drawn is disabled, making it impossible to later change e.g. the
        background colour. To force the axes patch to be displayed while still
        hiding the axes, set the ``draw_frame`` keyword argument to ``True``.
        """
        if field is None:
            field = self.fields
        for f in iter_fields(field):
            self.plots[f].hide_axes(draw_frame=draw_frame)
        return self

    def show_axes(self, field=None):
        """
        Shows the axes for a plot and updates the size of the
        plot accordingly.  Defaults to operating on all fields for a
        PlotContainer object.  See hide_axes().

        Parameters
        ----------

        field : string, field tuple, or list of strings or field tuples (optional)
            The name of the field(s) that we want to show the axes.
        """
        if field is None:
            field = self.fields
        for f in iter_fields(field):
            self.plots[f].show_axes()
        return self


class ImagePlotContainer(PlotContainer, abc.ABC):
    """A container for plots with colorbars."""

    _colorbar_valid = False

    def __init__(self, data_source, figure_size, fontsize):
        super().__init__(data_source, figure_size, fontsize)
        self._callbacks = []
        self._colorbar_label = PlotDictionary(self.data_source, lambda: None)

    def _get_default_handlers(
        self, field, default_display_units: Unit
    ) -> tuple[NormHandler, ColorbarHandler]:
        usr_units_str = get_default_from_config(
            self.data_source, field=field, keys="units", defaults=[None]
        )
        if usr_units_str is not None:
            usr_units = Unit(usr_units_str)
            d1 = usr_units.dimensions
            d2 = default_display_units.dimensions

            if d1 == d2:
                display_units = usr_units
            elif getattr(self, "projected", False) and d2 / d1 == length:
                path_length_units = Unit(
                    ytcfg.get_most_specific(
                        "plot", *field, "path_length_units", fallback="cm"
                    ),
                    registry=self.data_source.ds.unit_registry,
                )
                display_units = usr_units * path_length_units
            else:
                raise YTConfigurationError(
                    f"Invalid units in configuration file for field {field!r}. "
                    f"Found {usr_units!r}"
                )
        else:
            display_units = default_display_units

        pnh = NormHandler(self.data_source, display_units=display_units)

        cbh = ColorbarHandler(
            cmap=get_default_from_config(
                self.data_source,
                field=field,
                keys="cmap",
                defaults=[None],
            )
        )
        return pnh, cbh

    @accepts_all_fields
    @invalidate_plot
    def set_cmap(self, field, cmap):
        """set the colormap for one of the fields

        Parameters
        ----------
        field : string
            the field to set the colormap
            if field == 'all', applies to all plots.
        cmap : string or tuple
            If a string, will be interpreted as name of the colormap.
            If a tuple, it is assumed to be of the form (name, type, number)
            to be used for palettable functionality. (name, type, number, bool)
            can be used to specify if a reverse colormap is to be used.

        """
        self._colorbar_valid = False
        self.plots[field].colorbar_handler.cmap = cmap
        return self

    @accepts_all_fields
    @invalidate_plot
    def set_background_color(self, field, color=None):
        """set the background color to match provided color

        Parameters
        ----------
        field : string
            the field to set the colormap
            if field == 'all', applies to all plots.
        color : string or RGBA tuple (optional)
            if set, set the background color to this color
            if unset, background color is set to the bottom value of
            the color map

        """
        cbh = self[field].colorbar_handler
        cbh.background_color = color
        return self

    @accepts_all_fields
    @invalidate_plot
    def set_zlim(
        self,
        field,
        zmin: float | Quantity | Literal["min"] | Unset = UNSET,
        zmax: float | Quantity | Literal["max"] | Unset = UNSET,
        dynamic_range: float | None = None,
    ):
        """set the scale of the colormap

        Parameters
        ----------
        field : string
            the field to set a colormap scale
            if field == 'all', applies to all plots.
        zmin : float, Quantity, or 'min'
            the new minimum of the colormap scale. If 'min', will
            set to the minimum value in the current view.
        zmax : float, Quantity, or 'max'
            the new maximum of the colormap scale. If 'max', will
            set to the maximum value in the current view.

        Other Parameters
        ----------------
        dynamic_range : float (default: None)
            The dynamic range of the image.
            If zmin == None, will set zmin = zmax / dynamic_range
            If zmax == None, will set zmax = zmin * dynamic_range

        """
        if zmin is UNSET and zmax is UNSET:
            raise TypeError("Missing required argument zmin or zmax")

        if zmin is UNSET:
            zmin = None
        elif zmin is None:
            # this sentinel value juggling is barely maintainable
            # this use case is deprecated so we can simplify the logic here
            # in the future and use `None` as the default value,
            # instead of the custom sentinel UNSET
            issue_deprecation_warning(
                "Passing `zmin=None` explicitly is deprecated. "
                "If you wish to explicitly set zmin to the minimal "
                "data value, pass `zmin='min'` instead. "
                "Otherwise leave this argument unset.",
                since="4.1",
                stacklevel=5,
            )
            zmin = "min"

        if zmax is UNSET:
            zmax = None
        elif zmax is None:
            # see above
            issue_deprecation_warning(
                "Passing `zmax=None` explicitly is deprecated. "
                "If you wish to explicitly set zmax to the maximal "
                "data value, pass `zmax='max'` instead. "
                "Otherwise leave this argument unset.",
                since="4.1",
                stacklevel=5,
            )
            zmax = "max"

        pnh = self.plots[field].norm_handler
        pnh.vmin = zmin
        pnh.vmax = zmax
        pnh.dynamic_range = dynamic_range

        return self

    @accepts_all_fields
    @invalidate_plot
    def set_colorbar_minorticks(self, field, state):
        """turn colorbar minor ticks on or off in the current plot

        Displaying minor ticks reduces performance; turn them off
        using set_colorbar_minorticks('all', False) if drawing speed is a problem.

        Parameters
        ----------
        field : string
            the field to remove colorbar minorticks
            if field == 'all', applies to all plots.
        state : bool
            the state indicating 'on' (True) or 'off' (False)
        """
        self.plots[field].colorbar_handler.draw_minorticks = state
        return self

    @invalidate_plot
    def set_colorbar_label(self, field, label):
        r"""
        Sets the colorbar label.

        Parameters
        ----------
        field : str or tuple
          The name of the field to modify the label for.
        label : str
          The new label

        >>> plot.set_colorbar_label(
        ...     ("gas", "density"), "Dark Matter Density (g cm$^{-3}$)"
        ... )

        """
        field = self.data_source._determine_fields(field)
        self._colorbar_label[field] = label
        return self

    def _get_axes_labels(self, field):
        return (self._xlabel, self._ylabel, self._colorbar_label[field])


class BaseLinePlot(PlotContainer, abc.ABC):
    # A common ancestor to LinePlot and ProfilePlot

    @abc.abstractmethod
    def _get_axrect(self):
        pass

    def _get_plot_instance(self, field):
        if field in self.plots:
            return self.plots[field]
        axrect = self._get_axrect()

        pnh = NormHandler(
            self.data_source, display_units=self.data_source.ds.field_info[field].units
        )
        finfo = self.data_source.ds._get_field_info(field)
        if not finfo.take_log:
            pnh.norm_type = Normalize
        plot = PlotMPL(self.figure_size, axrect, norm_handler=pnh)
        self.plots[field] = plot

        return plot



"""
AdaptaHOP frontend tests



"""

import numpy as np
import pytest

from yt.frontends.adaptahop.data_structures import AdaptaHOPDataset
from yt.testing import requires_file
from yt.utilities.answer_testing.framework import data_dir_load

r0 = "output_00080/info_00080.txt"
brick_old = "output_00080_halos/tree_bricks080"
brick_new = "output_00080_new_halos/tree_bricks080"


@requires_file(r0)
@requires_file(brick_old)
@requires_file(brick_new)
@pytest.mark.parametrize("brick", [brick_old, brick_new])
def test_opening(brick):
    ds_parent = data_dir_load(r0)
    ds = data_dir_load(brick, kwargs={"parent_ds": ds_parent})

    ds.index

    assert isinstance(ds, AdaptaHOPDataset)


@requires_file(r0)
@requires_file(brick_old)
@requires_file(brick_new)
@pytest.mark.parametrize("brick", [brick_old, brick_new])
def test_field_access(brick):
    ds_parent = data_dir_load(r0)
    ds = data_dir_load(brick, kwargs={"parent_ds": ds_parent})

    skip_list = ("particle_identities", "mesh_id", "radius_profile", "rho_profile")
    fields = [
        (ptype, field)
        for (ptype, field) in ds.derived_field_list
        if (ptype == "halos") and (field not in skip_list)
    ]

    ad = ds.all_data()

    for ptype, field in fields:
        ad[ptype, field]


@requires_file(r0)
@requires_file(brick_old)
@requires_file(brick_new)
@pytest.mark.parametrize("brick", [brick_old, brick_new])
def test_get_halo(brick):
    ds_parent = data_dir_load(r0)
    ds = data_dir_load(brick, kwargs={"parent_ds": ds_parent})

    halo = ds.halo(1, ptype="io")

    # Check halo object has position, velocity, mass and members attributes
    for attr_name in ("mass", "position", "velocity", "member_ids"):
        getattr(halo, attr_name)

    members = np.sort(halo.member_ids)

    # Check sphere contains all the members
    id_sphere = halo.sphere["io", "particle_identity"].astype("int64")
    assert len(np.lib.arraysetops.setdiff1d(members, id_sphere)) == 0

    # Check region contains *only* halo particles
    id_reg = np.sort(halo["io", "particle_identity"].astype("int64"))

    np.testing.assert_equal(members, id_reg)



import numpy as np
from numpy.testing import assert_array_less

from yt.testing import fake_random_ds


def setup_module():
    from yt.config import ytcfg

    ytcfg["yt", "log_level"] = 50
    ytcfg["yt", "internals", "within_testing"] = True


def _difference(x1, x2, dw):
    rel = x1 - x2
    rel[rel > dw / 2.0] -= dw
    rel[rel < -dw / 2.0] += dw
    return rel


def test_ellipsoid():
    # We decompose in different ways
    cs = [
        np.array([0.5, 0.5, 0.5]),
        np.array([0.1, 0.2, 0.3]),
        np.array([0.8, 0.8, 0.8]),
    ]
    np.random.seed(0x4D3D3D3)
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(64, nprocs=nprocs)
        DW = ds.domain_right_edge - ds.domain_left_edge
        min_dx = 2.0 / ds.domain_dimensions
        ABC = np.random.random((3, 12)) * 0.1
        e0s = np.random.random((3, 12))
        tilts = np.random.random(12)
        ABC[:, 0] = 0.1
        for i in range(12):
            for c in cs:
                A, B, C = sorted(ABC[:, i], reverse=True)
                A = max(A, min_dx[0])
                B = max(B, min_dx[1])
                C = max(C, min_dx[2])
                e0 = e0s[:, i]
                tilt = tilts[i]
                ell = ds.ellipsoid(c, A, B, C, e0, tilt)
                assert_array_less(ell["index", "radius"], A)
                p = np.array([ell["index", ax] for ax in "xyz"])
                dot_evec = [np.zeros_like(ell["index", "radius"]) for i in range(3)]
                vecs = [ell._e0, ell._e1, ell._e2]
                mags = [ell._A, ell._B, ell._C]
                my_c = np.array([c] * p.shape[1]).transpose()
                dot_evec = [de.to_ndarray() for de in dot_evec]
                mags = [m.to_ndarray() for m in mags]
                for ax_i in range(3):
                    dist = _difference(p[ax_i, :], my_c[ax_i, :], DW[ax_i])
                    for ax_j in range(3):
                        dot_evec[ax_j] += dist * vecs[ax_j][ax_i]
                dist = 0
                for ax_i in range(3):
                    dist += dot_evec[ax_i] ** 2.0 / mags[ax_i] ** 2.0
                assert_array_less(dist, 1.0)
